# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import itertools

import pytest

from numba.cuda.core import config, errors
from numba.cuda import types, typing
from numba.cuda.typeconv import Conversion

from numba.cuda.testing import CUDATestCase
from numba.cuda.tests.cudapy.test_typeconv import CompatibilityTestMixin
from numba.cuda.core.untyped_passes import TranslateByteCode, IRProcessing
from numba.cuda.core.typed_passes import PartialTypeInference
from numba.cuda.core.compiler_machinery import FunctionPass, register_pass
import unittest

from numba.cuda.flags import Flags


i8 = types.int8
i16 = types.int16
i32 = types.int32
i64 = types.int64
u8 = types.uint8
u16 = types.uint16
u32 = types.uint32
u64 = types.uint64
f32 = types.float32
f64 = types.float64
c64 = types.complex64
c128 = types.complex128


class TestUnify(CUDATestCase):
    """
    Tests for type unification with a typing context.
    """

    int_unify = {
        ("uint8", "uint8"): "uint8",
        ("int8", "int8"): "int8",
        ("uint16", "uint16"): "uint16",
        ("int16", "int16"): "int16",
        ("uint32", "uint32"): "uint32",
        ("int32", "int32"): "int32",
        ("uint64", "uint64"): "uint64",
        ("int64", "int64"): "int64",
        ("int8", "uint8"): "int16",
        ("int8", "uint16"): "int32",
        ("int8", "uint32"): "int64",
        ("uint8", "int32"): "int32",
        ("uint8", "uint64"): "uint64",
        ("int16", "int8"): "int16",
        ("int16", "uint8"): "int16",
        ("int16", "uint16"): "int32",
        ("int16", "uint32"): "int64",
        ("int16", "int64"): "int64",
        ("int16", "uint64"): "float64",
        ("uint16", "uint8"): "uint16",
        ("uint16", "uint32"): "uint32",
        ("uint16", "int32"): "int32",
        ("uint16", "uint64"): "uint64",
        ("int32", "int8"): "int32",
        ("int32", "int16"): "int32",
        ("int32", "uint32"): "int64",
        ("int32", "int64"): "int64",
        ("uint32", "uint8"): "uint32",
        ("uint32", "int64"): "int64",
        ("uint32", "uint64"): "uint64",
        ("int64", "int8"): "int64",
        ("int64", "uint8"): "int64",
        ("int64", "uint16"): "int64",
        ("uint64", "int8"): "float64",
        ("uint64", "int32"): "float64",
        ("uint64", "int64"): "float64",
    }

    def assert_unify(self, aty, bty, expected):
        ctx = typing.Context()
        template = "{0}, {1} -> {2} != {3}"
        for unify_func in ctx.unify_types, ctx.unify_pairs:
            unified = unify_func(aty, bty)
            self.assertEqual(
                unified,
                expected,
                msg=template.format(aty, bty, unified, expected),
            )
            unified = unify_func(bty, aty)
            self.assertEqual(
                unified,
                expected,
                msg=template.format(bty, aty, unified, expected),
            )

    def assert_unify_failure(self, aty, bty):
        self.assert_unify(aty, bty, None)

    def test_integer(self):
        ctx = typing.Context()  # noqa: F841
        for aty, bty in itertools.product(
            types.integer_domain, types.integer_domain
        ):
            key = (str(aty), str(bty))
            try:
                expected = self.int_unify[key]
            except KeyError:
                expected = self.int_unify[key[::-1]]
            self.assert_unify(aty, bty, getattr(types, expected))

    def test_bool(self):
        aty = types.boolean
        for bty in types.integer_domain:
            self.assert_unify(aty, bty, bty)
        # Not sure about this one, but it respects transitivity
        for cty in types.real_domain:
            self.assert_unify(aty, cty, cty)

    def unify_number_pair_test(self, n):
        """
        Test all permutations of N-combinations of numeric types and ensure
        that the order of types in the sequence is irrelevant.
        """
        ctx = typing.Context()
        for tys in itertools.combinations(types.number_domain, n):
            res = [
                ctx.unify_types(*comb) for comb in itertools.permutations(tys)
            ]
            first_result = res[0]
            # Sanity check
            self.assertIsInstance(first_result, types.Number)
            # All results must be equal
            for other in res[1:]:
                self.assertEqual(first_result, other)

    def test_unify_number_pair(self):
        self.unify_number_pair_test(2)
        self.unify_number_pair_test(3)

    def test_none_to_optional(self):
        """
        Test unification of `none` and multiple number types to optional type
        """
        ctx = typing.Context()
        for tys in itertools.combinations(types.number_domain, 2):
            # First unify without none, to provide the control value
            tys = list(tys)
            expected = types.Optional(ctx.unify_types(*tys))
            results = [
                ctx.unify_types(*comb)
                for comb in itertools.permutations(tys + [types.none])
            ]
            # All results must be equal
            for res in results:
                self.assertEqual(res, expected)

    def test_none(self):
        aty = types.none
        bty = types.none
        self.assert_unify(aty, bty, types.none)

    def test_optional(self):
        aty = types.Optional(i32)
        bty = types.none
        self.assert_unify(aty, bty, aty)
        aty = types.Optional(i32)
        bty = types.Optional(i64)
        self.assert_unify(aty, bty, bty)
        aty = types.Optional(i32)
        bty = i64
        self.assert_unify(aty, bty, types.Optional(i64))
        # Failure
        aty = types.Optional(i32)
        bty = types.Optional(types.slice3_type)
        self.assert_unify_failure(aty, bty)

    def test_tuple(self):
        aty = types.UniTuple(i32, 3)
        bty = types.UniTuple(i64, 3)
        self.assert_unify(aty, bty, types.UniTuple(i64, 3))
        # (Tuple, UniTuple) -> Tuple
        aty = types.UniTuple(i32, 2)
        bty = types.Tuple((i16, i64))
        self.assert_unify(aty, bty, types.Tuple((i32, i64)))
        aty = types.UniTuple(i64, 0)
        bty = types.Tuple(())
        self.assert_unify(aty, bty, bty)
        # (Tuple, Tuple) -> Tuple
        aty = types.Tuple((i8, i16, i32))
        bty = types.Tuple((i32, i16, i8))
        self.assert_unify(aty, bty, types.Tuple((i32, i16, i32)))
        aty = types.Tuple((i8, i32))
        bty = types.Tuple((i32, i8))
        self.assert_unify(aty, bty, types.Tuple((i32, i32)))
        aty = types.Tuple((i8, i16))
        bty = types.Tuple((i16, i8))
        self.assert_unify(aty, bty, types.Tuple((i16, i16)))
        # Different number kinds
        aty = types.UniTuple(f64, 3)
        bty = types.UniTuple(c64, 3)
        self.assert_unify(aty, bty, types.UniTuple(c128, 3))
        # Tuples of tuples
        aty = types.UniTuple(types.Tuple((u32, f32)), 2)
        bty = types.UniTuple(types.Tuple((i16, f32)), 2)
        self.assert_unify(aty, bty, types.UniTuple(types.Tuple((i64, f32)), 2))
        # Failures
        aty = types.UniTuple(i32, 1)
        bty = types.UniTuple(types.slice3_type, 1)
        self.assert_unify_failure(aty, bty)
        aty = types.UniTuple(i32, 1)
        bty = types.UniTuple(i32, 2)
        self.assert_unify_failure(aty, bty)
        aty = types.Tuple((i8, types.slice3_type))
        bty = types.Tuple((i32, i8))
        self.assert_unify_failure(aty, bty)

    def test_optional_tuple(self):
        # Unify to optional tuple
        aty = types.none
        bty = types.UniTuple(i32, 2)
        self.assert_unify(aty, bty, types.Optional(types.UniTuple(i32, 2)))
        aty = types.Optional(types.UniTuple(i16, 2))
        bty = types.UniTuple(i32, 2)
        self.assert_unify(aty, bty, types.Optional(types.UniTuple(i32, 2)))
        # Unify to tuple of optionals
        aty = types.Tuple((types.none, i32))
        bty = types.Tuple((i16, types.none))
        self.assert_unify(
            aty, bty, types.Tuple((types.Optional(i16), types.Optional(i32)))
        )
        aty = types.Tuple((types.Optional(i32), i64))
        bty = types.Tuple((i16, types.Optional(i8)))
        self.assert_unify(
            aty, bty, types.Tuple((types.Optional(i32), types.Optional(i64)))
        )

    def test_arrays(self):
        aty = types.Array(i32, 3, "C")
        bty = types.Array(i32, 3, "A")
        self.assert_unify(aty, bty, bty)
        aty = types.Array(i32, 3, "C")
        bty = types.Array(i32, 3, "F")
        self.assert_unify(aty, bty, types.Array(i32, 3, "A"))
        aty = types.Array(i32, 3, "C")
        bty = types.Array(i32, 3, "C", readonly=True)
        self.assert_unify(aty, bty, bty)
        aty = types.Array(i32, 3, "A")
        bty = types.Array(i32, 3, "C", readonly=True)
        self.assert_unify(aty, bty, types.Array(i32, 3, "A", readonly=True))
        # Failures
        aty = types.Array(i32, 2, "C")
        bty = types.Array(i32, 3, "C")
        self.assert_unify_failure(aty, bty)
        aty = types.Array(i32, 2, "C")
        bty = types.Array(u32, 2, "C")
        self.assert_unify_failure(aty, bty)

    def test_list(self):
        aty = types.List(types.undefined)
        bty = types.List(i32)
        self.assert_unify(aty, bty, bty)
        aty = types.List(i16)
        bty = types.List(i32)
        self.assert_unify(aty, bty, bty)
        aty = types.List(types.Tuple([i32, i16]))
        bty = types.List(types.Tuple([i16, i64]))
        cty = types.List(types.Tuple([i32, i64]))
        self.assert_unify(aty, bty, cty)
        # Different reflections
        aty = types.List(i16, reflected=True)
        bty = types.List(i32)
        cty = types.List(i32, reflected=True)
        self.assert_unify(aty, bty, cty)
        # Incompatible dtypes
        aty = types.List(i16)
        bty = types.List(types.Tuple([i16]))
        self.assert_unify_failure(aty, bty)

    def test_set(self):
        # Different reflections
        aty = types.Set(i16, reflected=True)
        bty = types.Set(i32)
        cty = types.Set(i32, reflected=True)
        self.assert_unify(aty, bty, cty)
        # Incompatible dtypes
        aty = types.Set(i16)
        bty = types.Set(types.Tuple([i16]))
        self.assert_unify_failure(aty, bty)

    def test_range(self):
        aty = types.range_state32_type
        bty = types.range_state64_type
        self.assert_unify(aty, bty, bty)


class TestTypeConversion(CompatibilityTestMixin, unittest.TestCase):
    """
    Test for conversion between types with a typing context.
    """

    def assert_can_convert(self, aty, bty, expected):
        ctx = typing.Context()
        got = ctx.can_convert(aty, bty)
        self.assertEqual(got, expected)

    def assert_cannot_convert(self, aty, bty):
        ctx = typing.Context()
        got = ctx.can_convert(aty, bty)
        self.assertIsNone(got)

    def test_convert_number_types(self):
        # Check that Context.can_convert() is compatible with the default
        # number conversion rules registered in the typeconv module
        # (which is used internally by the C _Dispatcher object).
        ctx = typing.Context()
        self.check_number_compatibility(ctx.can_convert)

    def test_tuple(self):
        # UniTuple -> UniTuple
        aty = types.UniTuple(i32, 3)
        bty = types.UniTuple(i64, 3)
        self.assert_can_convert(aty, aty, Conversion.exact)
        self.assert_can_convert(aty, bty, Conversion.promote)
        aty = types.UniTuple(i32, 3)
        bty = types.UniTuple(f64, 3)
        self.assert_can_convert(aty, bty, Conversion.safe)
        # Tuple -> Tuple
        aty = types.Tuple((i32, i32))
        bty = types.Tuple((i32, i64))
        self.assert_can_convert(aty, bty, Conversion.promote)
        # UniTuple <-> Tuple
        aty = types.UniTuple(i32, 2)
        bty = types.Tuple((i32, i64))
        self.assert_can_convert(aty, bty, Conversion.promote)
        self.assert_can_convert(bty, aty, Conversion.unsafe)
        # Empty tuples
        aty = types.UniTuple(i64, 0)
        bty = types.UniTuple(i32, 0)
        cty = types.Tuple(())
        self.assert_can_convert(aty, bty, Conversion.safe)
        self.assert_can_convert(bty, aty, Conversion.safe)
        self.assert_can_convert(aty, cty, Conversion.safe)
        self.assert_can_convert(cty, aty, Conversion.safe)
        # Failures
        aty = types.UniTuple(i64, 3)
        bty = types.UniTuple(types.none, 3)
        self.assert_cannot_convert(aty, bty)
        aty = types.UniTuple(i64, 2)
        bty = types.UniTuple(i64, 3)

    def test_arrays(self):
        # Different layouts
        aty = types.Array(i32, 3, "C")
        bty = types.Array(i32, 3, "A")
        self.assert_can_convert(aty, bty, Conversion.safe)
        aty = types.Array(i32, 2, "C")
        bty = types.Array(i32, 2, "F")
        self.assert_cannot_convert(aty, bty)
        # Different mutabilities
        aty = types.Array(i32, 3, "C")
        bty = types.Array(i32, 3, "C", readonly=True)
        self.assert_can_convert(aty, aty, Conversion.exact)
        self.assert_can_convert(bty, bty, Conversion.exact)
        self.assert_can_convert(aty, bty, Conversion.safe)
        self.assert_cannot_convert(bty, aty)
        # Various failures
        aty = types.Array(i32, 2, "C")
        bty = types.Array(i32, 3, "C")
        self.assert_cannot_convert(aty, bty)
        aty = types.Array(i32, 2, "C")
        bty = types.Array(i64, 2, "C")
        self.assert_cannot_convert(aty, bty)

    def test_optional(self):
        aty = types.int32
        bty = types.Optional(i32)
        self.assert_can_convert(types.none, bty, Conversion.promote)
        self.assert_can_convert(aty, bty, Conversion.promote)
        self.assert_cannot_convert(bty, types.none)
        self.assert_can_convert(bty, aty, Conversion.safe)  # XXX ???
        # Optional array
        aty = types.Array(i32, 2, "C")
        bty = types.Optional(aty)
        self.assert_can_convert(types.none, bty, Conversion.promote)
        self.assert_can_convert(aty, bty, Conversion.promote)
        self.assert_can_convert(bty, aty, Conversion.safe)
        aty = types.Array(i32, 2, "C")
        bty = types.Optional(aty.copy(layout="A"))
        self.assert_can_convert(aty, bty, Conversion.safe)  # C -> A
        self.assert_cannot_convert(bty, aty)  # A -> C
        aty = types.Array(i32, 2, "C")
        bty = types.Optional(aty.copy(layout="F"))
        self.assert_cannot_convert(aty, bty)
        self.assert_cannot_convert(bty, aty)


class TestResolveOverload(unittest.TestCase):
    """
    Tests for typing.Context.resolve_overload().
    """

    def assert_resolve_overload(self, cases, args, expected):
        ctx = typing.Context()
        got = ctx.resolve_overload("foo", cases, args, {})
        self.assertEqual(got, expected)

    def test_non_ambiguous_match(self):
        def check(args, expected):
            self.assert_resolve_overload(cases, args, expected)
            # Order shouldn't matter here
            self.assert_resolve_overload(cases[::-1], args, expected)

        cases = [i8(i8, i8), i32(i32, i32), f64(f64, f64)]
        # Exact match
        check((i8, i8), cases[0])
        check((i32, i32), cases[1])
        check((f64, f64), cases[2])
        # "Promote" conversion
        check((i8, i16), cases[1])
        check((i32, i8), cases[1])
        check((i32, i8), cases[1])
        check((f32, f32), cases[2])
        # "Safe" conversion
        check((u32, u32), cases[2])
        # "Unsafe" conversion
        check((i64, i64), cases[2])

    def test_ambiguous_match(self):
        # When the best match is ambiguous (there is a tie), the first
        # best case in original sequence order should be returned.
        def check(args, expected, expected_reverse):
            self.assert_resolve_overload(cases, args, expected)
            self.assert_resolve_overload(cases[::-1], args, expected_reverse)

        cases = [i16(i16, i16), i32(i32, i32), f64(f64, f64)]
        # Two "promote" conversions
        check((i8, i8), cases[0], cases[1])
        # Two "safe" conversions
        check((u16, u16), cases[1], cases[2])

        cases = [i32(i32, i32), f32(f32, f32)]
        # Two "unsafe" conversions
        check((u32, u32), cases[0], cases[1])

    def test_ambiguous_error(self):
        ctx = typing.Context()
        cases = [i16(i16, i16), i32(i32, i32)]
        with self.assertRaises(TypeError) as raises:
            ctx.resolve_overload(
                "foo", cases, (i8, i8), {}, allow_ambiguous=False
            )
        self.assertEqual(
            str(raises.exception).splitlines(),
            [
                "Ambiguous overloading for foo (int8, int8):",
                "(int16, int16) -> int16",
                "(int32, int32) -> int32",
            ],
        )


@register_pass(mutates_CFG=False, analysis_only=True)
class DummyCR(FunctionPass):
    """Dummy pass to add "cr" to compiler state to avoid errors in TyperCompiler since
    it doesn't have lowering.
    """

    _name = "dummy_cr"

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        state.cr = 1  # arbitrary non-None value
        return True


def get_func_typing_errs(func, arg_types):
    """
    Get typing errors for function 'func'. It creates a pipeline that runs untyped
    passes as well as type inference.
    """
    from numba.cuda.compiler import CompilerBase, PassManager

    class TyperCompiler(CompilerBase):
        """A compiler pipeline that skips passes after typing (provides partial typing info
        but not lowering).
        """

        def define_pipelines(self):
            pm = PassManager("custom_pipeline")
            pm.add_pass(TranslateByteCode, "analyzing bytecode")
            pm.add_pass(IRProcessing, "processing IR")
            pm.add_pass(PartialTypeInference, "do partial typing")
            pm.add_pass_after(DummyCR, PartialTypeInference)
            pm.finalize()
            return [pm]

    from numba.cuda.descriptor import cuda_target

    typingctx = cuda_target.typing_context
    targetctx = cuda_target.target_context
    library = None
    return_type = None
    _locals = {}
    flags = Flags()

    pipeline = TyperCompiler(
        typingctx, targetctx, library, arg_types, return_type, flags, _locals
    )
    pipeline.compile_extra(func)
    return pipeline.state.typing_errors


@pytest.mark.skipif(config.ENABLE_CUDASIM, reason="not enable in the simulator")
def test_partial_typing_error():
    """
    Make sure partial typing stores type errors in compiler state properly
    """

    # example with type unification error
    def impl(flag):
        if flag:
            a = 1
        else:
            a = ""
        return a

    with pytest.raises(
        errors.TypingError,
        match=r"Cannot unify Literal\[int\]\(1\) and Literal\[str\]\(\) for 'a'",
    ):
        get_func_typing_errs(impl, (types.bool_,))


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import pytest
import concurrent.futures
import multiprocessing as mp
import os
import itertools
import pickle

import numpy as np

from numba import cuda
from numba.cuda.testing import (
    skip_on_arm,
    skip_on_cudasim,
    skip_under_cuda_memcheck,
    skip_on_wsl2,
    CUDATestCase,
    ForeignArray,
)
from numba.cuda.tests.support import linux_only, windows_only
import unittest


FUTURE_TIMEOUT = 4


def base_ipc_handle_test(handle, size, parent_pid):
    pid = os.getpid()
    assert pid != parent_pid
    dtype = np.dtype(np.intp)
    with cuda.open_ipc_array(
        handle, shape=size // dtype.itemsize, dtype=dtype
    ) as darr:
        # copy the data to host
        return darr.copy_to_host()


def serialize_ipc_handle_test(handle, parent_pid):
    pid = os.getpid()
    assert pid != parent_pid

    dtype = np.dtype(np.intp)
    darr = handle.open_array(
        cuda.current_context(),
        shape=handle.size // dtype.itemsize,
        dtype=dtype,
    )
    # copy the data to host
    arr = darr.copy_to_host()
    handle.close()
    return arr


def ipc_array_test(ipcarr, parent_pid):
    pid = os.getpid()
    assert pid != parent_pid
    with ipcarr as darr:
        arr = darr.copy_to_host()
        with pytest.raises(ValueError, match="IpcHandle is already opened"):
            with ipcarr:
                pass
    return arr


class CUDAIpcTestCase(CUDATestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.exe = concurrent.futures.ProcessPoolExecutor(
            mp_context=mp.get_context("spawn")
        )

    @classmethod
    def tearDownClass(cls) -> None:
        cls.exe.shutdown()
        del cls.exe


@linux_only
@skip_under_cuda_memcheck("Hangs cuda-memcheck")
@skip_on_cudasim("Ipc not available in CUDASIM")
@skip_on_arm("CUDA IPC not supported on ARM in Numba")
@skip_on_wsl2("CUDA IPC unreliable on WSL2; skipping IPC tests")
class TestIpcMemory(CUDAIpcTestCase):
    def test_ipc_handle(self):
        # prepare data for IPC
        arr = np.arange(10, dtype=np.intp)
        devarr = cuda.to_device(arr)

        # create IPC handle
        ctx = cuda.current_context()
        ipch = ctx.get_ipc_handle(devarr.gpu_data)

        # manually prepare for serialization as bytes
        handle_bytes = ipch.handle.reserved
        size = ipch.size

        # spawn new process for testing
        fut = self.exe.submit(
            base_ipc_handle_test, handle_bytes, size, parent_pid=os.getpid()
        )
        out = fut.result(timeout=FUTURE_TIMEOUT)
        np.testing.assert_equal(arr, out)

    def variants(self):
        # Test with no slicing and various different slices
        indices = (None, slice(3, None), slice(3, 8), slice(None, 8))
        # Test with a Numba DeviceNDArray, or an array from elsewhere through
        # the CUDA Array Interface
        foreigns = (False, True)
        return itertools.product(indices, foreigns)

    def check_ipc_handle_serialization(self, index_arg=None, foreign=False):
        # prepare data for IPC
        arr = np.arange(10, dtype=np.intp)
        devarr = cuda.to_device(arr)
        if index_arg is not None:
            devarr = devarr[index_arg]
        if foreign:
            devarr = cuda.as_cuda_array(ForeignArray(devarr))
        expect = devarr.copy_to_host()

        # create IPC handle
        ctx = cuda.current_context()
        ipch = ctx.get_ipc_handle(devarr.gpu_data)

        # pickle
        buf = pickle.dumps(ipch)
        ipch_recon = pickle.loads(buf)
        self.assertIs(ipch_recon.base, None)
        self.assertEqual(ipch_recon.size, ipch.size)

        self.assertEqual(ipch_recon.handle.reserved, ipch.handle.reserved)

        # spawn new process for testing
        fut = self.exe.submit(
            serialize_ipc_handle_test, ipch, parent_pid=os.getpid()
        )
        out = fut.result(timeout=FUTURE_TIMEOUT)
        np.testing.assert_equal(expect, out)

    def test_ipc_handle_serialization(self):
        for (
            index,
            foreign,
        ) in self.variants():
            with self.subTest(index=index, foreign=foreign):
                self.check_ipc_handle_serialization(index, foreign)

    def check_ipc_array(self, index_arg=None, foreign=False):
        # prepare data for IPC
        arr = np.arange(10, dtype=np.intp)
        devarr = cuda.to_device(arr)
        # Slice
        if index_arg is not None:
            devarr = devarr[index_arg]
        if foreign:
            devarr = cuda.as_cuda_array(ForeignArray(devarr))
        expect = devarr.copy_to_host()
        ipch = devarr.get_ipc_handle()

        # spawn new process for testing
        fut = self.exe.submit(ipc_array_test, ipch, parent_pid=os.getpid())
        out = fut.result(timeout=FUTURE_TIMEOUT)
        np.testing.assert_equal(expect, out)

    def test_ipc_array(self):
        for (
            index,
            foreign,
        ) in self.variants():
            with self.subTest(index=index, foreign=foreign):
                self.check_ipc_array(index, foreign)


def staged_ipc_handle_test(handle, device_num, parent_pid):
    pid = os.getpid()
    assert pid != parent_pid
    with cuda.gpus[device_num]:
        this_ctx = cuda.devices.get_context()
        deviceptr = handle.open_staged(this_ctx)
        arrsize = handle.size // np.dtype(np.intp).itemsize
        hostarray = np.zeros(arrsize, dtype=np.intp)
        cuda.driver.device_to_host(
            hostarray,
            deviceptr,
            size=handle.size,
        )
        handle.close()
        return hostarray


def staged_ipc_array_test(ipcarr, device_num, parent_pid):
    pid = os.getpid()
    assert pid != parent_pid
    with cuda.gpus[device_num]:
        with ipcarr as darr:
            arr = darr.copy_to_host()
            with pytest.raises(ValueError, match="IpcHandle is already opened"):
                with ipcarr:
                    pass
    return arr


@linux_only
@skip_under_cuda_memcheck("Hangs cuda-memcheck")
@skip_on_cudasim("Ipc not available in CUDASIM")
@skip_on_arm("CUDA IPC not supported on ARM in Numba")
@skip_on_wsl2("CUDA IPC unreliable on WSL2; skipping IPC tests")
class TestIpcStaged(CUDAIpcTestCase):
    def test_staged(self):
        # prepare data for IPC
        arr = np.arange(10, dtype=np.intp)
        devarr = cuda.to_device(arr)
        # create IPC handle
        ctx = cuda.current_context()
        ipch = ctx.get_ipc_handle(devarr.gpu_data)
        # pickle
        buf = pickle.dumps(ipch)
        ipch_recon = pickle.loads(buf)
        self.assertIs(ipch_recon.base, None)
        self.assertEqual(ipch_recon.handle.reserved, ipch.handle.reserved)
        self.assertEqual(ipch_recon.size, ipch.size)

        # Test on every CUDA devices
        ngpus = len(cuda.gpus)

        for device_num in range(ngpus):
            future = self.exe.submit(
                staged_ipc_handle_test, ipch, device_num, parent_pid=os.getpid()
            )
            np.testing.assert_equal(arr, future.result(timeout=FUTURE_TIMEOUT))

    def test_ipc_array(self):
        for device_num in range(len(cuda.gpus)):
            # prepare data for IPC
            arr = np.random.random(10)
            devarr = cuda.to_device(arr)
            ipch = devarr.get_ipc_handle()

            # spawn new process for testing
            fut = self.exe.submit(
                staged_ipc_array_test, ipch, device_num, parent_pid=os.getpid()
            )
            out = fut.result(timeout=FUTURE_TIMEOUT)
            np.testing.assert_equal(arr, out)


@windows_only
@skip_on_cudasim("Ipc not available in CUDASIM")
class TestIpcNotSupported(CUDATestCase):
    def test_unsupported(self):
        arr = np.arange(10, dtype=np.intp)
        devarr = cuda.to_device(arr)
        with self.assertRaises(OSError) as raises:
            devarr.get_ipc_handle()
        errmsg = str(raises.exception)
        self.assertIn("OS does not support CUDA IPC", errmsg)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from collections import namedtuple
from numba.cuda.tests.support import override_config, captured_stdout
from numba.cuda.testing import skip_on_cudasim
from numba import cuda
from numba.cuda import types
from numba.cuda.testing import CUDATestCase
from numba.cuda.core import config
from textwrap import dedent
import math
import itertools
import re
import unittest
import warnings
from numba.cuda.core.errors import NumbaDebugInfoWarning
from numba.cuda.tests.support import ignore_internal_warnings
import numpy as np
import inspect


@skip_on_cudasim("Simulator does not produce debug dumps")
class TestCudaDebugInfo(CUDATestCase):
    """
    These tests only checks the compiled PTX for debuginfo section
    """

    def _getasm(self, fn, sig):
        fn.compile(sig)
        return fn.inspect_asm(sig)

    def _check(self, fn, sig, expect):
        asm = self._getasm(fn, sig=sig)
        re_section_dbginfo = re.compile(r"\.section\s+\.debug_info\s+{")
        match = re_section_dbginfo.search(asm)
        assertfn = self.assertIsNotNone if expect else self.assertIsNone
        assertfn(match, msg=asm)

    def test_no_debuginfo_in_asm(self):
        @cuda.jit(debug=False, opt=False)
        def foo(x):
            x[0] = 1

        self._check(foo, sig=(types.int32[:],), expect=False)

    def test_debuginfo_in_asm(self):
        @cuda.jit(debug=True, opt=False)
        def foo(x):
            x[0] = 1

        self._check(foo, sig=(types.int32[:],), expect=True)

    def test_environment_override(self):
        with override_config("CUDA_DEBUGINFO_DEFAULT", 1):
            # Using default value
            @cuda.jit(opt=False)
            def foo(x):
                x[0] = 1

            self._check(foo, sig=(types.int32[:],), expect=True)

            # User override default value
            @cuda.jit(debug=False)
            def bar(x):
                x[0] = 1

            self._check(bar, sig=(types.int32[:],), expect=False)

    def test_issue_5835(self):
        # Invalid debug metadata would segfault NVVM when any function was
        # compiled with debug turned on and optimization off. This eager
        # compilation should not crash anything.
        @cuda.jit((types.int32[::1],), debug=True, opt=False)
        def f(x):
            x[0] = 0

    def test_issue_9888(self):
        # Compiler created symbol should not be emitted in DILocalVariable
        # See Numba Issue #9888 https://github.com/numba/numba/pull/9888
        sig = (types.boolean,)

        @cuda.jit(sig, debug=True, opt=False)
        def f(cond):
            if cond:
                x = 1  # noqa: F841
            else:
                x = 0  # noqa: F841

        llvm_ir = f.inspect_llvm(sig)
        # A varible name starting with "bool" in the debug metadata
        pat = r"!DILocalVariable\(.*name:\s+\"bool"
        match = re.compile(pat).search(llvm_ir)
        self.assertIsNone(match, msg=llvm_ir)

    def test_bool_type(self):
        sig = (types.int32, types.int32)

        @cuda.jit("void(int32, int32)", debug=True, opt=False)
        def f(x, y):
            z = x == y  # noqa: F841

        llvm_ir = f.inspect_llvm(sig)

        # extract the metadata node id from `type` field of DILocalVariable
        pat = r'!DILocalVariable\(.*name:\s+"z".*type:\s+!(\d+)'
        match = re.compile(pat).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)
        mdnode_id = match.group(1)

        # verify the DIBasicType has correct encoding attribute DW_ATE_boolean
        pat = rf"!{mdnode_id}\s+=\s+!DIBasicType\(.*DW_ATE_boolean"
        match = re.compile(pat).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)

    def test_grid_group_type(self):
        sig = (types.int32,)

        @cuda.jit(sig, debug=True, opt=False)
        def f(x):
            grid = cuda.cg.this_grid()  # noqa: F841

        llvm_ir = f.inspect_llvm(sig)

        pat = r'!DIBasicType\(.*DW_ATE_unsigned, name: "GridGroup", size: 64'
        match = re.compile(pat).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)

    @unittest.skip("Wrappers no longer exist")
    def test_wrapper_has_debuginfo(self):
        sig = (types.int32[::1],)

        @cuda.jit(sig, debug=True, opt=0)
        def f(x):
            x[0] = 1

        llvm_ir = f.inspect_llvm(sig)

        defines = [
            line
            for line in llvm_ir.splitlines()
            if 'define void @"_ZN6cudapy' in line
        ]

        # Make sure we only found one definition
        self.assertEqual(len(defines), 1)

        wrapper_define = defines[0]
        self.assertIn("!dbg", wrapper_define)

    def test_debug_function_calls_internal_impl(self):
        # Calling a function in a module generated from an implementation
        # internal to Numba requires multiple modules to be compiled with NVVM -
        # the internal implementation, and the caller. This example uses two
        # modules because the `in (2, 3)` is implemented with:
        #
        # numba::cpython::listobj::in_seq::$3clocals$3e::seq_contains_impl$242(
        #     UniTuple<long long, 2>,
        #     int
        # )
        #
        # This is condensed from this reproducer in Issue 5311:
        # https://github.com/numba/numba/issues/5311#issuecomment-674206587

        @cuda.jit((types.int32[:], types.int32[:]), debug=True, opt=False)
        def f(inp, outp):
            outp[0] = 1 if inp[0] in (2, 3) else 3

    def test_debug_function_calls_device_function(self):
        # Calling a device function requires compilation of multiple modules
        # with NVVM - one for the caller and one for the callee. This checks
        # that we don't cause an NVVM error in this case.

        @cuda.jit(device=True, debug=True, opt=0)
        def threadid():
            return cuda.blockDim.x * cuda.blockIdx.x + cuda.threadIdx.x

        @cuda.jit((types.int32[:],), debug=True, opt=0)
        def kernel(arr):
            i = cuda.grid(1)
            if i < len(arr):
                arr[i] = threadid()

    def _test_chained_device_function(self, kernel_debug, f1_debug, f2_debug):
        @cuda.jit(device=True, debug=f2_debug, opt=False)
        def f2(x):
            return x + 1

        @cuda.jit(device=True, debug=f1_debug, opt=False)
        def f1(x, y):
            return x - f2(y)

        @cuda.jit((types.int32, types.int32), debug=kernel_debug, opt=False)
        def kernel(x, y):
            f1(x, y)

        kernel[1, 1](1, 2)

    def test_chained_device_function(self):
        # Calling a device function that calls another device function from a
        # kernel with should succeed regardless of which jit decorators have
        # debug=True. See Issue #7159.

        debug_opts = itertools.product(*[(True, False)] * 3)

        for kernel_debug, f1_debug, f2_debug in debug_opts:
            with self.subTest(
                kernel_debug=kernel_debug, f1_debug=f1_debug, f2_debug=f2_debug
            ):
                self._test_chained_device_function(
                    kernel_debug, f1_debug, f2_debug
                )

    def _test_chained_device_function_two_calls(
        self, kernel_debug, f1_debug, f2_debug
    ):
        @cuda.jit(device=True, debug=f2_debug, opt=False)
        def f2(x):
            return x + 1

        @cuda.jit(device=True, debug=f1_debug, opt=False)
        def f1(x, y):
            return x - f2(y)

        @cuda.jit(debug=kernel_debug, opt=False)
        def kernel(x, y):
            f1(x, y)
            f2(x)

        kernel[1, 1](1, 2)

    def test_chained_device_function_two_calls(self):
        # Calling a device function that calls a leaf device function from a
        # kernel, and calling the leaf device function from the kernel should
        # succeed, regardless of which jit decorators have debug=True. See
        # Issue #7159.

        debug_opts = itertools.product(*[(True, False)] * 3)

        for kernel_debug, f1_debug, f2_debug in debug_opts:
            with self.subTest(
                kernel_debug=kernel_debug, f1_debug=f1_debug, f2_debug=f2_debug
            ):
                self._test_chained_device_function_two_calls(
                    kernel_debug, f1_debug, f2_debug
                )

    def test_chained_device_three_functions(self):
        # Like test_chained_device_function, but with enough functions (three)
        # to ensure that the recursion visits all the way down the call tree
        # when fixing linkage of functions for debug.
        def three_device_fns(kernel_debug, leaf_debug):
            @cuda.jit(device=True, debug=leaf_debug, opt=False)
            def f3(x):
                return x * x

            @cuda.jit(device=True)
            def f2(x):
                return f3(x) + 1

            @cuda.jit(device=True)
            def f1(x, y):
                return x - f2(y)

            @cuda.jit(debug=kernel_debug, opt=False)
            def kernel(x, y):
                f1(x, y)

            kernel[1, 1](1, 2)

        # Check when debug on the kernel, on the leaf, and not on any function.
        three_device_fns(kernel_debug=True, leaf_debug=True)
        three_device_fns(kernel_debug=True, leaf_debug=False)
        three_device_fns(kernel_debug=False, leaf_debug=True)
        three_device_fns(kernel_debug=False, leaf_debug=False)

    def _test_kernel_args_types(self):
        sig = (types.int32, types.int32)

        @cuda.jit("void(int32, int32)", debug=True, opt=False)
        def f(x, y):
            z = x + y  # noqa: F841

        llvm_ir = f.inspect_llvm(sig)

        # extract the metadata node id from `types` field of DISubroutineType
        pat = r"!DISubroutineType\(types:\s+!(\d+)\)"
        match = re.compile(pat).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)
        mdnode_id = match.group(1)

        # extract the metadata node ids from the flexible node of types
        pat = rf"!{mdnode_id}\s+=\s+!{{\s+!(\d+),\s+!(\d+)\s+}}"
        match = re.compile(pat).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)
        mdnode_id1 = match.group(1)
        mdnode_id2 = match.group(2)

        # verify each of the two metadata nodes match expected type
        pat = rf'!{mdnode_id1}\s+=\s+!DIBasicType\(.*DW_ATE_signed,\s+name:\s+"int32"'  # noqa: E501
        match = re.compile(pat).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)
        pat = rf'!{mdnode_id2}\s+=\s+!DIBasicType\(.*DW_ATE_signed,\s+name:\s+"int32"'  # noqa: E501
        match = re.compile(pat).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)

    def test_kernel_args_types(self):
        self._test_kernel_args_types()

    def test_kernel_args_types_dump(self):
        # see issue#135
        with override_config("DUMP_LLVM", 1):
            with captured_stdout():
                self._test_kernel_args_types()

    def test_kernel_args_names(self):
        sig = (types.int32,)

        @cuda.jit("void(int32)", debug=True, opt=False)
        def f(x):
            z = x  # noqa: F841

        llvm_ir = f.inspect_llvm(sig)

        # Verify argument name is not prefixed with "arg."
        pat = r"define void @.*\(i32 %\"x\"\)"
        match = re.compile(pat).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)
        pat = r"define void @.*\(i32 %\"arg\.x\"\)"
        match = re.compile(pat).search(llvm_ir)
        self.assertIsNone(match, msg=llvm_ir)

    def test_llvm_dbg_value(self):
        sig = (types.int32, types.int32)

        @cuda.jit("void(int32, int32)", debug=True, opt=False)
        def f(x, y):
            z1 = x  # noqa: F841
            z2 = 100  # noqa: F841
            z3 = y  # noqa: F841
            z4 = True  # noqa: F841

        llvm_ir = f.inspect_llvm(sig)
        # Verify the call to llvm.dbg.declare is replaced by llvm.dbg.value
        pat1 = r'call void @"llvm.dbg.declare"'
        match = re.compile(pat1).search(llvm_ir)
        self.assertIsNone(match, msg=llvm_ir)
        pat2 = r'call void @"llvm.dbg.value"'
        match = re.compile(pat2).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)

    def test_llvm_dbg_value_range(self):
        sig = (types.int64,)

        @cuda.jit("void(int64,)", debug=True, opt=False)
        def foo(x):
            """
            CHECK: store i1 true, i1* %"second.1"
            CHECK: call void @"llvm.dbg.value"
            CHECK: store i1 true, i1* %"second.2"
            CHECK: call void @"llvm.dbg.value"

            CHECK: %[[VAL_1:.*]] = load i1, i1* %"second.2"
            CHECK: %[[VAL_2:.*]] = load i1, i1* %[[VAL_3:.*]]
            CHECK: store i1 %[[VAL_1]], i1* %[[VAL_3]]
            CHECK: call void @"llvm.dbg.value"(metadata i1 %[[VAL_1]], metadata ![[VAL_4:[0-9]+]]

            CHECK: ![[VAL_4]] = !DILocalVariable{{.+}}name: "second"
            """
            if x > 0:
                second = x > 10
            else:
                second = True
            if second:
                pass

        ir = foo.inspect_llvm()[sig]
        self.assertFileCheckMatches(ir, foo.__doc__)

    def test_no_user_var_alias(self):
        sig = (types.int32, types.int32)

        @cuda.jit("void(int32, int32)", debug=True, opt=False)
        def f(x, y):
            z = x  # noqa: F841
            z = y  # noqa: F841

        llvm_ir = f.inspect_llvm(sig)
        pat = r'!DILocalVariable.*name:\s+"z\$1".*'
        match = re.compile(pat).search(llvm_ir)
        self.assertIsNone(match, msg=llvm_ir)

    def test_no_literal_type(self):
        sig = (types.int32,)

        @cuda.jit("void(int32)", debug=True, opt=False)
        def f(x):
            z = x  # noqa: F841
            z = 100  # noqa: F841
            z = True  # noqa: F841

        llvm_ir = f.inspect_llvm(sig)
        pat = r'!DIBasicType.*name:\s+"Literal.*'
        match = re.compile(pat).search(llvm_ir)
        self.assertIsNone(match, msg=llvm_ir)

    @unittest.skipIf(
        config.CUDA_DEBUG_POLY, "Uses old union format, not variant_part"
    )
    def test_union_poly_types(self):
        sig = (types.int32, types.int32)

        @cuda.jit("void(int32, int32)", debug=True, opt=False)
        def f(x, y):
            foo = 100  # noqa: F841
            foo = 2.34  # noqa: F841
            foo = True  # noqa: F841
            foo = 200  # noqa: F841

        llvm_ir = f.inspect_llvm(sig)
        # Extract the type node id
        pat1 = r'!DILocalVariable\(.*name: "foo".*type: !(\d+)\)'
        match = re.compile(pat1).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)
        mdnode_id = match.group(1)
        # Verify the union type and extract the elements node id
        pat2 = rf"!{mdnode_id} = distinct !DICompositeType\(elements: !(\d+),.*size: 64, tag: DW_TAG_union_type\)"  # noqa: E501
        match = re.compile(pat2).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)
        mdnode_id = match.group(1)
        # Extract the member node ids
        pat3 = r"!{ !(\d+), !(\d+), !(\d+) }"
        match = re.compile(pat3).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)
        mdnode_id1 = match.group(1)
        mdnode_id2 = match.group(2)
        mdnode_id3 = match.group(3)
        # Verify the member nodes
        pat4 = rf'!{mdnode_id1} = !DIDerivedType(.*name: "_bool", size: 8, tag: DW_TAG_member)'  # noqa: E501
        match = re.compile(pat4).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)
        pat5 = rf'!{mdnode_id2} = !DIDerivedType(.*name: "_float64", size: 64, tag: DW_TAG_member)'  # noqa: E501
        match = re.compile(pat5).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)
        pat6 = rf'!{mdnode_id3} = !DIDerivedType(.*name: "_int64", size: 64, tag: DW_TAG_member)'  # noqa: E501
        match = re.compile(pat6).search(llvm_ir)
        self.assertIsNotNone(match, msg=llvm_ir)

    def test_union_debug(self):
        @cuda.jit("void(u8, int64[::1])", debug=True, opt=False)
        def a_union_use_case(arg, results):
            foo = 1
            foo = arg
            if foo < 1:
                foo = 2
                return
            bar = foo == 0
            results[0] = 1 if not bar else 0

        with captured_stdout() as out:
            results = cuda.to_device(np.zeros(16, dtype=np.int64))
            a_union_use_case[1, 1](100, results)
            print(results.copy_to_host())
        expected = "[1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0]"
        self.assertIn(expected, out.getvalue())

    @unittest.skipUnless(config.CUDA_DEBUG_POLY, "CUDA_DEBUG_POLY not enabled")
    def test_poly_variant_part(self):
        """Test polymorphic variables with DW_TAG_variant_part.

        This test verifies that when CUDA_DEBUG_POLY is enabled,
        polymorphic variables generate proper DWARF5 variant_part
        debug information with discriminator and variant members.
        """
        # Typed constant: i8 0, i8 1, etc. | Node reference: !123, !456
        if config.CUDA_DEBUG_POLY_USE_TYPED_CONST:
            extradata_pattern = "i8 {{[0-9]+}}"
        else:
            extradata_pattern = "{{![0-9]+}}"

        @cuda.jit("void()", debug=True, opt=False)
        def f():
            foo = 100  # noqa: F841
            foo = 3.14  # noqa: F841
            foo = True  # noqa: F841
            foo = np.int32(42)  # noqa: F841

        llvm_ir = f.inspect_llvm()[tuple()]

        # Build FileCheck pattern dynamically based on config
        # Capture node IDs and verify the hierarchical structure
        check_pattern = """
            CHECK-DAG: !DILocalVariable({{.*}}name: "foo"{{.*}}type: [[WRAPPER:![0-9]+]]
            CHECK-DAG: [[WRAPPER]] = !DICompositeType({{.*}}elements: [[ELEMENTS:![0-9]+]]{{.*}}name: "variant_wrapper_struct"{{.*}}size: 128{{.*}}tag: DW_TAG_structure_type)
            CHECK-DAG: [[ELEMENTS]] = !{ [[DISC:![0-9]+]], [[VPART:![0-9]+]] }
            CHECK-DAG: [[DISC]] = !DIDerivedType({{.*}}name: "discriminator-{{[0-9]+}}"{{.*}}size: 8{{.*}}tag: DW_TAG_member)
            CHECK-DAG: [[VPART]] = !DICompositeType({{.*}}discriminator: [[DISC]]{{.*}}elements: [[VMEMBERS:![0-9]+]]{{.*}}tag: DW_TAG_variant_part)
            CHECK-DAG: [[VMEMBERS]] = !{ [[VM1:![0-9]+]], [[VM2:![0-9]+]], [[VM3:![0-9]+]], [[VM4:![0-9]+]] }
            CHECK-DAG: [[VM1]] = !DIDerivedType({{.*}}extraData: EXTRADATA{{.*}}name: "_bool"{{.*}}offset: 8{{.*}}tag: DW_TAG_member)
            CHECK-DAG: [[VM2]] = !DIDerivedType({{.*}}extraData: EXTRADATA{{.*}}name: "_float64"{{.*}}offset: 64{{.*}}tag: DW_TAG_member)
            CHECK-DAG: [[VM3]] = !DIDerivedType({{.*}}extraData: EXTRADATA{{.*}}name: "_int32"{{.*}}offset: 32{{.*}}tag: DW_TAG_member)
            CHECK-DAG: [[VM4]] = !DIDerivedType({{.*}}extraData: EXTRADATA{{.*}}name: "_int64"{{.*}}offset: 64{{.*}}tag: DW_TAG_member)
        """.replace("EXTRADATA", extradata_pattern)

        self.assertFileCheckMatches(llvm_ir, check_pattern)

    def test_DW_LANG(self):
        @cuda.jit(debug=True, opt=False)
        def foo():
            """
            CHECK: distinct !DICompileUnit
            CHECK-SAME: emissionKind: FullDebug
            CHECK-SAME: isOptimized: true
            CHECK-SAME: language: DW_LANG_C_plus_plus
            CHECK-SAME: producer: "clang (Numba)"
            """
            pass

        foo[1, 1]()

        llvm_ir = foo.inspect_llvm()[tuple()]
        self.assertFileCheckMatches(llvm_ir, foo.__doc__)

    def test_DILocation(self):
        """Tests that DILocation information is reasonable.

        The kernel `foo` produces LLVM like:
        define function() {
        entry:
          alloca
          store 0 to alloca
          <arithmetic for doing the operations on b, c, d>
          setup for print
          branch
        other_labels:
        ... <elided>
        }

        The following checks that:
        * the alloca and store have no !dbg
        * the arithmetic occurs in the order defined and with !dbg
        * that the !dbg entries are monotonically increasing in value with
          source line number
        """
        sig = (types.float64,)

        @cuda.jit(sig, debug=True, opt=False)
        def foo(a):
            """
            CHECK-LABEL: define void @{{.+}}foo
            CHECK: entry:

            CHECK: %[[VAL_0:.*]] = alloca double
            CHECK-NOT: !dbg
            CHECK: store double 0.0, double* %[[VAL_0]]
            CHECK-NOT: !dbg
            CHECK: %[[VAL_1:.*]] = alloca double
            CHECK-NOT: !dbg
            CHECK: store double 0.0, double* %[[VAL_1]]
            CHECK-NOT: !dbg
            CHECK: %[[VAL_2:.*]] = alloca double
            CHECK-NOT: !dbg
            CHECK: store double 0.0, double* %[[VAL_2]]
            CHECK-NOT: !dbg
            CHECK: %[[VAL_3:.*]] = alloca double
            CHECK-NOT: !dbg
            CHECK: store double 0.0, double* %[[VAL_3]]
            CHECK-NOT: !dbg
            CHECK: %[[VAL_4:.*]] = alloca double
            CHECK-NOT: !dbg
            CHECK: store double 0.0, double* %[[VAL_4]]
            CHECK-NOT: !dbg
            CHECK: %[[VAL_5:.*]] = alloca double
            CHECK-NOT: !dbg
            CHECK: store double 0.0, double* %[[VAL_5]]
            CHECK-NOT: !dbg
            CHECK: %[[VAL_6:.*]] = alloca i8*
            CHECK-NOT: !dbg
            CHECK: store i8* null, i8** %[[VAL_6]]
            CHECK-NOT: !dbg
            CHECK: %[[VAL_7:.*]] = alloca i8*
            CHECK-NOT: !dbg
            CHECK: store i8* null, i8** %[[VAL_7]]
            CHECK-NOT: !dbg

            CHECK: br label %"[[ENTRY:.+]]"
            CHECK-NOT: !dbg
            CHECK: [[ENTRY]]:

            CHECK: fadd{{.+}} !dbg ![[DBGADD:[0-9]+]]
            CHECK: fmul{{.+}} !dbg ![[DBGMUL:[0-9]+]]
            CHECK: fdiv{{.+}} !dbg ![[DBGDIV:[0-9]+]]

            CHECK: ![[DBGADD]] = !DILocation
            CHECK: ![[DBGMUL]] = !DILocation
            CHECK: ![[DBGDIV]] = !DILocation
            """
            b = a + 1.23
            c = b * 2.34
            a = b / c

        ir = foo.inspect_llvm()[sig]
        self.assertFileCheckMatches(ir, foo.__doc__)

    def test_missing_source(self):
        strsrc = """
        def foo():
            pass
        """
        l = dict()
        exec(dedent(strsrc), {}, l)
        foo = cuda.jit(debug=True, opt=False)(l["foo"])

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", NumbaDebugInfoWarning)
            ignore_internal_warnings()
            foo[1, 1]()

        self.assertEqual(len(w), 1)
        found = w[0]
        self.assertEqual(found.category, NumbaDebugInfoWarning)
        msg = str(found.message)
        # make sure the warning contains the right message
        self.assertIn("Could not find source for function", msg)
        # and refers to the offending function
        self.assertIn(str(foo.py_func), msg)

    def test_no_if_op_bools_declared(self):
        @cuda.jit(
            "int64(boolean, boolean)",
            debug=True,
            opt=False,
            _dbg_optnone=True,
            device=True,
        )
        def choice(cond1, cond2):
            """
            CHECK: define void @{{.+}}choices
            """
            if cond1 and cond2:
                return 1
            else:
                return 2

        ir_content = choice.inspect_llvm()[choice.signatures[0]]
        # We should not declare variables used as the condition in if ops.
        # See Numba PR #9888: https://github.com/numba/numba/pull/9888

        for line in ir_content.splitlines():
            if "llvm.dbg.declare" in line:
                self.assertNotIn("bool", line)

    def test_llvm_inliner_flag_conflict(self):
        # bar will be marked as 'alwaysinline', but when DEBUGINFO_DEFAULT is
        # set functions are not marked as 'alwaysinline' and this results in a
        # conflict. baz will not be marked as 'alwaysinline' as a result of
        # DEBUGINFO_DEFAULT

        @cuda.jit(forceinline=True)
        def bar(x):
            return math.sin(x)

        @cuda.jit(forceinline=False)
        def baz(x):
            return math.cos(x)

        @cuda.jit(opt=True)
        def foo(x, y):
            """
            CHECK-LABEL: define void @{{.+}}foo
            CHECK: call i32 @"[[BAR:.+]]"(
            CHECK: call i32 @"[[BAZ:.+]]"(

            CHECK-DAG: declare i32 @"[[BAR]]"({{.+}}alwaysinline
            CHECK-DAG: declare i32 @"[[BAZ]]"(
            CHECK-DAG: define linkonce_odr i32 @"[[BAR]]"({{.+}}alwaysinline
            CHECK-DAG: define linkonce_odr i32 @"[[BAZ]]"(
            """
            a = bar(y)
            b = baz(y)
            x[0] = a + b

        # check it compiles
        with override_config("DEBUGINFO_DEFAULT", 1):
            result = cuda.device_array(1, dtype=np.float32)
            foo[1, 1](result, np.pi)
            result.copy_to_host()

        result_host = math.sin(np.pi) + math.cos(np.pi)
        self.assertPreciseEqual(result[0], result_host)

        ir_content = foo.inspect_llvm()[foo.signatures[0]]
        self.assertFileCheckMatches(ir_content, foo.__doc__)

        # Check that the device functions call the appropriate device
        # math functions and have the correct attributes.
        self.assertFileCheckMatches(
            ir_content,
            """
            CHECK: define linkonce_odr i32 @{{.+}}bar
            CHECK-SAME: alwaysinline
            CHECK-NEXT: {
            CHECK-NEXT: {{.*}}:
            CHECK-NEXT: br label {{.*}}
            CHECK-NEXT: {{.*}}:
            CHECK-NEXT: call double @"__nv_sin"
            CHECK-NEXT: store double {{.*}}, double* {{.*}}
            CHECK-NEXT: ret i32 0
            CHECK-NEXT: }
        """,
        )

        self.assertFileCheckMatches(
            ir_content,
            """
            CHECK: define linkonce_odr i32 @{{.+}}baz
            CHECK-NOT: alwaysinline
            CHECK-NEXT: {
            CHECK-NEXT: {{.*}}:
            CHECK-NEXT: br label {{.*}}
            CHECK-NEXT: {{.*}}:
            CHECK-NEXT: call double @"__nv_cos"
            CHECK-NEXT: store double {{.*}}, double* {{.*}}
            CHECK-NEXT: ret i32 0
            CHECK-NEXT: }
        """,
        )

    def test_DILocation_versioned_variables(self):
        """Tests that DILocation information for versions of variables matches
        up to their definition site."""

        @cuda.jit(debug=True, opt=False)
        def foo(dest, n):
            """
            CHECK: define void @{{.+}}foo
            CHECK: store i64 5, i64* %"c{{.+}} !dbg ![[STORE5:.+]]
            CHECK: store i64 1, i64* %"c{{.+}} !dbg ![[STORE1:.+]]
            CHECK: [[STORE5]] = !DILocation(
            CHECK: [[STORE1]] = !DILocation(
            """
            if n:
                c = 5
            else:
                c = 1
            dest[0] = c

        foo_source_lines, foo_source_lineno = inspect.getsourcelines(
            foo.py_func
        )

        result = cuda.device_array(1, dtype=np.int32)
        foo[1, 1](result, 1)
        result.copy_to_host()
        self.assertEqual(result[0], 5)

        ir_content = foo.inspect_llvm()[foo.signatures[0]]
        self.assertFileCheckMatches(ir_content, foo.__doc__)

        # Collect lines pertaining to the function `foo` and debuginfo
        # metadata
        lines = ir_content.splitlines()
        debuginfo_equals = re.compile(r"!(\d+) = ")
        debug_info_lines = list(
            filter(lambda x: debuginfo_equals.search(x), lines)
        )

        function_start_regex = re.compile(r"define void @.+foo")
        function_start_lines = list(
            filter(
                lambda x: function_start_regex.search(x[1]), enumerate(lines)
            )
        )
        function_end_lines = list(
            filter(lambda x: x[1] == "}", enumerate(lines))
        )
        foo_ir_lines = lines[
            function_start_lines[0][0] : function_end_lines[0][0]
        ]

        # Check the if condition's debuginfo
        cond_branch = list(filter(lambda x: "br i1" in x, foo_ir_lines))
        self.assertEqual(len(cond_branch), 1)
        self.assertIn("!dbg", cond_branch[0])
        cond_branch_dbginfo_node = cond_branch[0].split("!dbg")[1].strip()
        cond_branch_dbginfos = list(
            filter(
                lambda x: cond_branch_dbginfo_node + " = " in x,
                debug_info_lines,
            )
        )
        self.assertEqual(len(cond_branch_dbginfos), 1)
        cond_branch_dbginfo = cond_branch_dbginfos[0]

        # Check debuginfo for the store instructions
        store_1_lines = list(filter(lambda x: "store i64 1" in x, foo_ir_lines))
        store_5_lines = list(filter(lambda x: "store i64 5" in x, foo_ir_lines))

        self.assertEqual(len(store_1_lines), 2)
        self.assertEqual(len(store_5_lines), 2)

        store_1_dbginfo_set = set(
            map(lambda x: x.split("!dbg")[1].strip(), store_1_lines)
        )
        store_5_dbginfo_set = set(
            map(lambda x: x.split("!dbg")[1].strip(), store_5_lines)
        )
        self.assertEqual(len(store_1_dbginfo_set), 1)
        self.assertEqual(len(store_5_dbginfo_set), 1)
        store_1_dbginfo_node = store_1_dbginfo_set.pop()
        store_5_dbginfo_node = store_5_dbginfo_set.pop()
        store_1_dbginfos = list(
            filter(
                lambda x: store_1_dbginfo_node + " = " in x, debug_info_lines
            )
        )
        store_5_dbginfos = list(
            filter(
                lambda x: store_5_dbginfo_node + " = " in x, debug_info_lines
            )
        )
        self.assertEqual(len(store_1_dbginfos), 1)
        self.assertEqual(len(store_5_dbginfos), 1)
        store_1_dbginfo = store_1_dbginfos[0]
        store_5_dbginfo = store_5_dbginfos[0]

        # Ensure the line numbers match what we expect based on the Python source
        line_number_regex = re.compile(r"line: (\d+)")
        LineNumbers = namedtuple(
            "LineNumbers", ["cond_branch", "store_5", "store_1"]
        )
        line_number_matches = LineNumbers(
            *map(
                lambda x: line_number_regex.search(x),
                [cond_branch_dbginfo, store_5_dbginfo, store_1_dbginfo],
            )
        )
        self.assertTrue(
            all(
                map(
                    lambda x: x is not None,
                    line_number_matches,
                )
            )
        )
        line_numbers = LineNumbers(
            *map(
                lambda x: int(x.group(1)),
                line_number_matches,
            )
        )
        source_line_numbers = LineNumbers(
            *map(
                lambda x: x[0] + foo_source_lineno,
                filter(
                    lambda x: "c = " in x[1] or "if n:" in x[1],
                    enumerate(foo_source_lines),
                ),
            )
        )
        self.assertEqual(line_numbers, source_line_numbers)

    def test_debuginfo_asm(self):
        def foo():
            pass

        foo_debug = cuda.jit(debug=True, opt=False)(foo)
        foo_debug[1, 1]()
        asm = foo_debug.inspect_asm()[foo_debug.signatures[0]]
        self.assertFileCheckMatches(
            asm,
            """
            CHECK: .section{{.+}}.debug
        """,
        )

        foo_nodebug = cuda.jit(debug=False)(foo)
        foo_nodebug[1, 1]()
        asm = foo_nodebug.inspect_asm()[foo_nodebug.signatures[0]]
        self.assertFileCheckMatches(
            asm,
            """
            CHECK-NOT: .section{{.+}}.debug
        """,
        )


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from collections import namedtuple, defaultdict
import operator
import warnings
from functools import partial

from llvmlite import ir as llvm_ir

from numba.cuda import HAS_NUMBA
from numba.cuda.core import ir
from numba.cuda import debuginfo, cgutils, utils, typing, types
from numba.cuda.core import (
    ir_utils,
    targetconfig,
    funcdesc,
    config,
    generators,
    removerefctpass,
)

from numba.cuda.core.errors import (
    LoweringError,
    new_error_context,
    TypingError,
    LiteralTypingError,
    UnsupportedError,
    NumbaDebugInfoWarning,
)
from numba.cuda.core.funcdesc import default_mangler
from numba.cuda.core.environment import Environment
from numba.cuda.core.analysis import compute_use_defs, must_use_alloca
from numba.cuda.misc.firstlinefinder import get_func_body_first_lineno
from numba.cuda.misc.coverage_support import get_registered_loc_notify


_VarArgItem = namedtuple("_VarArgItem", ("vararg", "index"))


class BaseLower(object):
    """
    Lower IR to LLVM
    """

    def __init__(self, context, library, fndesc, func_ir, metadata=None):
        self.library = library
        self.fndesc = fndesc
        self.blocks = dict(sorted(func_ir.blocks.items()))
        self.func_ir = func_ir
        self.generator_info = func_ir.generator_info
        self.metadata = metadata
        self.flags = targetconfig.ConfigStack.top_or_none()

        # Initialize LLVM
        self.module = self.library.create_ir_module(self.fndesc.unique_name)

        # Python execution environment (will be available to the compiled
        # function).
        self.env = Environment.from_fndesc(self.fndesc)

        # Internal states
        self.blkmap = {}
        self.pending_phis = {}
        self.varmap = {}
        self.firstblk = min(self.blocks.keys())
        self.loc = -1

        # Specializes the target context as seen inside the Lowerer
        # This adds:
        #  - environment: the python execution environment
        self.context = context.subtarget(
            environment=self.env, fndesc=self.fndesc
        )

        # Debuginfo
        dibuildercls = (
            self.context.DIBuilder
            if self.context.enable_debuginfo
            else debuginfo.DummyDIBuilder
        )

        # debuginfo def location
        self.defn_loc = self._compute_def_location()

        directives_only = self.flags.dbg_directives_only
        self.debuginfo = dibuildercls(
            module=self.module,
            filepath=func_ir.loc.filename,
            cgctx=context,
            directives_only=directives_only,
        )

        # Loc notify objects
        self._loc_notify_registry = get_registered_loc_notify()

        # Subclass initialization
        self.init()

    @property
    def call_conv(self):
        return self.context.call_conv

    def init(self):
        pass

    def init_pyapi(self):
        """
        Init the Python API and Environment Manager for the function being
        lowered.
        """
        if self.pyapi is not None:
            return
        self.pyapi = self.context.get_python_api(self.builder)

        # Store environment argument for later use
        self.env_manager = self.context.get_env_manager(self.builder)
        self.env_body = self.env_manager.env_body
        self.envarg = self.env_manager.env_ptr

    def _compute_def_location(self):
        # Debuginfo requires source to be accurate. Find it and warn if not
        # found. If it's not found, use the func_ir line + 1, this assumes that
        # the function definition is decorated with a 1 line jit decorator.
        defn_loc = self.func_ir.loc.with_lineno(self.func_ir.loc.line + 1)
        if self.context.enable_debuginfo:
            fn = self.func_ir.func_id.func
            optional_lno = get_func_body_first_lineno(fn)
            if optional_lno is not None:
                # -1 as lines start at 1 and this is an offset.
                offset = optional_lno - 1
                defn_loc = self.func_ir.loc.with_lineno(offset)
            else:
                msg = (
                    "Could not find source for function: "
                    f"{self.func_ir.func_id.func}. Debug line information "
                    "may be inaccurate."
                )
                warnings.warn(NumbaDebugInfoWarning(msg))
        return defn_loc

    def pre_lower(self):
        """
        Called before lowering all blocks.
        """
        # A given Lower object can be used for several LL functions
        # (for generators) and it's important to use a new API and
        # EnvironmentManager.
        self.pyapi = None
        self.debuginfo.mark_subprogram(
            function=self.builder.function,
            qualname=self.fndesc.qualname,
            argnames=self.fndesc.args,
            argtypes=self.fndesc.argtypes,
            line=self.defn_loc.line,
        )

        # When full debug info is enabled, disable inlining where possible, to
        # improve the quality of the debug experience. 'alwaysinline' functions
        # cannot have inlining disabled.
        attributes = self.builder.function.attributes
        full_debug = self.flags.debuginfo and not self.flags.dbg_directives_only
        if full_debug and "alwaysinline" not in attributes:
            attributes.add("noinline")

    def post_lower(self):
        """
        Called after all blocks are lowered
        """
        self.debuginfo.finalize()
        for notify in self._loc_notify_registry:
            notify.close()

    def pre_block(self, block):
        """
        Called before lowering a block.
        """

    def post_block(self, block):
        """
        Called after lowering a block.
        """

    def return_dynamic_exception(self, exc_class, exc_args, nb_types, loc=None):
        self.call_conv.return_dynamic_user_exc(
            self.builder,
            exc_class,
            exc_args,
            nb_types,
            loc=loc,
            func_name=self.func_ir.func_id.func_name,
        )

    def return_exception(self, exc_class, exc_args=None, loc=None):
        """Propagate exception to the caller."""
        self.call_conv.return_user_exc(
            self.builder,
            exc_class,
            exc_args,
            loc=loc,
            func_name=self.func_ir.func_id.func_name,
        )

    def set_exception(self, exc_class, exc_args=None, loc=None):
        """Set exception state in the current function."""
        self.call_conv.set_static_user_exc(
            self.builder,
            exc_class,
            exc_args,
            loc=loc,
            func_name=self.func_ir.func_id.func_name,
        )

    def emit_environment_object(self):
        """Emit a pointer to hold the Environment object."""
        # Define global for the environment and initialize it to NULL
        envname = self.context.get_env_name(self.fndesc)
        self.context.declare_env_global(self.module, envname)

    def lower(self):
        # Emit the Env into the module
        self.emit_environment_object()
        if self.generator_info is None:
            self.genlower = None
            self.lower_normal_function(self.fndesc)
        else:
            self.genlower = self.GeneratorLower(self)
            self.gentype = self.genlower.gentype

            self.genlower.lower_init_func(self)
            self.genlower.lower_next_func(self)
            if self.gentype.has_finalizer:
                self.genlower.lower_finalize_func(self)

        if config.DUMP_LLVM:
            utils.dump_llvm(self.fndesc, self.module)

        # Special optimization to remove NRT on functions that do not need it.
        if self.context.enable_nrt and self.generator_info is None:
            removerefctpass.remove_unnecessary_nrt_usage(
                self.function, context=self.context, fndesc=self.fndesc
            )

        # Run target specific post lowering transformation
        self.context.post_lowering(self.module, self.library)

        # Materialize LLVM Module
        self.library.add_ir_module(self.module)

    def extract_function_arguments(self):
        self.fnargs = self.call_conv.decode_arguments(
            self.builder, self.fndesc.argtypes, self.function
        )
        return self.fnargs

    def lower_normal_function(self, fndesc):
        """
        Lower non-generator *fndesc*.
        """
        self.setup_function(fndesc)

        # Init argument values
        self.extract_function_arguments()
        entry_block_tail = self.lower_function_body()

        # Close tail of entry block, do not emit debug metadata else the
        # unconditional jump gets associated with the metadata from the function
        # body end.
        with debuginfo.suspend_emission(self.builder):
            self.builder.position_at_end(entry_block_tail)
            self.builder.branch(self.blkmap[self.firstblk])

    def lower_function_body(self):
        """
        Lower the current function's body, and return the entry block.
        """
        # Init Python blocks
        for offset in self.blocks:
            bname = "B%s" % offset
            self.blkmap[offset] = self.function.append_basic_block(bname)

        self.pre_lower()
        # pre_lower() may have changed the current basic block
        entry_block_tail = self.builder.basic_block

        self.debug_print(
            "# function begin: {0}".format(self.fndesc.unique_name)
        )

        # Lower all blocks
        for offset, block in self.blocks.items():
            bb = self.blkmap[offset]
            self.builder.position_at_end(bb)
            self.debug_print(f"# lower block: {offset}")
            self.lower_block(block)
        self.post_lower()
        return entry_block_tail

    def lower_block(self, block):
        """
        Lower the given block.
        """
        self.pre_block(block)
        for inst in block.body:
            self.loc = inst.loc
            defaulterrcls = partial(LoweringError, loc=self.loc)
            with new_error_context(
                'lowering "{inst}" at {loc}',
                inst=inst,
                loc=self.loc,
                errcls_=defaulterrcls,
            ):
                self.lower_inst(inst)
        self.post_block(block)

    def create_cpython_wrapper(self, release_gil=False):
        """
        Create CPython wrapper(s) around this function (or generator).
        """
        if self.genlower:
            self.context.create_cpython_wrapper(
                self.library,
                self.genlower.gendesc,
                self.env,
                self.call_helper,
                release_gil=release_gil,
            )
        self.context.create_cpython_wrapper(
            self.library,
            self.fndesc,
            self.env,
            self.call_helper,
            release_gil=release_gil,
        )

    def create_cfunc_wrapper(self):
        """
        Create C wrapper around this function.
        """
        if self.genlower:
            raise UnsupportedError("generator as a first-class function type")
        self.context.create_cfunc_wrapper(
            self.library, self.fndesc, self.env, self.call_helper
        )

    def setup_function(self, fndesc):
        # Setup function
        self.function = self.context.declare_function(self.module, fndesc)
        if self.flags.dbg_optnone:
            attrset = self.function.attributes
            if "alwaysinline" not in attrset:
                attrset.add("optnone")
                attrset.add("noinline")
        self.entry_block = self.function.append_basic_block("entry")
        self.builder = llvm_ir.IRBuilder(self.entry_block)
        self.call_helper = self.call_conv.init_call_helper(self.builder)

    def typeof(self, varname):
        return self.fndesc.typemap[varname]

    def notify_loc(self, loc: ir.Loc) -> None:
        """Called when a new instruction with the given `loc` is about to be
        lowered.
        """
        for notify_obj in self._loc_notify_registry:
            notify_obj.notify(loc)

    def debug_print(self, msg):
        if config.DEBUG_JIT:
            self.context.debug_print(
                self.builder, f"DEBUGJIT [{self.fndesc.qualname}]: {msg}"
            )

    def print_variable(self, msg, varname):
        """Helper to emit ``print(msg, varname)`` for debugging.

        Parameters
        ----------
        msg : str
            Literal string to be printed.
        varname : str
            A variable name whose value will be printed.
        """
        argtys = (types.literal(msg), self.fndesc.typemap[varname])
        args = (
            self.context.get_dummy_value(),
            self.loadvar(varname),
        )
        sig = typing.signature(types.none, *argtys)

        impl = self.context.get_function(print, sig)
        impl(self.builder, args)


class Lower(BaseLower):
    GeneratorLower = generators.GeneratorLower

    def init(self):
        super().init()
        # find all singly assigned variables
        self._find_singly_assigned_variable()

    @property
    def _disable_sroa_like_opt(self):
        """Flags that the SROA like optimisation that Numba performs (which
        prevent alloca and subsequent load/store for locals) should be disabled.
        Currently, this is conditional solely on the presence of a request for
        the emission of debug information."""
        if self.flags is None:
            return False

        return self.flags.debuginfo and not self.flags.dbg_directives_only

    def _find_singly_assigned_variable(self):
        func_ir = self.func_ir
        blocks = func_ir.blocks

        sav = set()

        if not self.func_ir.func_id.is_generator:
            use_defs = compute_use_defs(blocks)
            alloca_vars = must_use_alloca(blocks)

            # Compute where variables are defined
            var_assign_map = defaultdict(set)
            for blk, vl in use_defs.defmap.items():
                for var in vl:
                    var_assign_map[var].add(blk)

            # Compute where variables are used
            var_use_map = defaultdict(set)
            for blk, vl in use_defs.usemap.items():
                for var in vl:
                    var_use_map[var].add(blk)

            # Keep only variables that are defined locally and used locally
            for var in var_assign_map:
                if var not in alloca_vars and len(var_assign_map[var]) == 1:
                    # Usemap does not keep locally defined variables.
                    if len(var_use_map[var]) == 0:
                        # Ensure that the variable is not defined multiple times
                        # in the block
                        [defblk] = var_assign_map[var]
                        assign_stmts = self.blocks[defblk].find_insts(ir.Assign)
                        assigns = [
                            stmt
                            for stmt in assign_stmts
                            if stmt.target.name == var
                        ]
                        if len(assigns) == 1:
                            sav.add(var)

        self._singly_assigned_vars = sav
        self._blk_local_varmap = {}

    def pre_block(self, block):
        from numba.cuda.core.unsafe import eh

        super(Lower, self).pre_block(block)
        self._cur_ir_block = block

        if block == self.firstblk:
            # create slots for all the vars, irrespective of whether they are
            # initialized, SSA will pick this up and warn users about using
            # uninitialized variables. Slots are added as alloca in the first
            # block
            bb = self.blkmap[self.firstblk]
            self.builder.position_at_end(bb)
            all_names = set()
            for block in self.blocks.values():
                for x in block.find_insts(ir.Del):
                    if x.value not in all_names:
                        all_names.add(x.value)
            for name in all_names:
                fetype = self.typeof(name)
                self._alloca_var(name, fetype)

        # Detect if we are in a TRY block by looking for a call to
        # `eh.exception_check`.
        for call in block.find_exprs(op="call"):
            defn = ir_utils.guard(
                ir_utils.get_definition,
                self.func_ir,
                call.func,
            )
            if defn is not None and isinstance(defn, ir.Global):
                if defn.value is eh.exception_check:
                    if isinstance(block.terminator, ir.Branch):
                        targetblk = self.blkmap[block.terminator.truebr]
                        # NOTE: This hacks in an attribute for call_conv to
                        #       pick up. This hack is no longer needed when
                        #       all old-style implementations are gone.
                        self.builder._in_try_block = {"target": targetblk}
                        break

    def post_block(self, block):
        # Clean-up
        try:
            del self.builder._in_try_block
        except AttributeError:
            pass

    def lower_inst(self, inst):
        # Set debug location for all subsequent LL instructions
        self.debuginfo.mark_location(self.builder, self.loc.line)
        self.notify_loc(self.loc)
        self.debug_print(str(inst))
        if isinstance(inst, ir.Assign):
            ty = self.typeof(inst.target.name)
            val = self.lower_assign(ty, inst)
            argidx = None
            # If this is a store from an arg, like x = arg.x then tell debuginfo
            # that this is the arg
            if isinstance(inst.value, ir.Arg):
                # NOTE: debug location is the `def <func>` line
                self.debuginfo.mark_location(self.builder, self.defn_loc.line)
                argidx = inst.value.index + 1  # args start at 1
            self.storevar(val, inst.target.name, argidx=argidx)

        elif isinstance(inst, ir.Branch):
            cond = self.loadvar(inst.cond.name)
            tr = self.blkmap[inst.truebr]
            fl = self.blkmap[inst.falsebr]

            condty = self.typeof(inst.cond.name)
            pred = self.context.cast(self.builder, cond, condty, types.boolean)
            assert pred.type == llvm_ir.IntType(1), (
                "cond is not i1: %s" % pred.type
            )
            self.builder.cbranch(pred, tr, fl)

        elif isinstance(inst, ir.Jump):
            target = self.blkmap[inst.target]
            self.builder.branch(target)

        elif isinstance(inst, ir.Return):
            if self.generator_info:
                # StopIteration
                self.genlower.return_from_generator(self)
                return
            val = self.loadvar(inst.value.name)
            oty = self.typeof(inst.value.name)
            ty = self.fndesc.restype
            if isinstance(ty, types.Optional):
                # If returning an optional type
                self.call_conv.return_optional_value(self.builder, ty, oty, val)
                return
            assert ty == oty, (
                "type '{}' does not match return type '{}'".format(oty, ty)
            )
            retval = self.context.get_return_value(self.builder, ty, val)
            self.call_conv.return_value(self.builder, retval)

        elif isinstance(inst, ir.PopBlock):
            pass  # this is just a marker

        elif isinstance(inst, ir.StaticSetItem):
            signature = self.fndesc.calltypes[inst]
            assert signature is not None
            try:
                impl = self.context.get_function("static_setitem", signature)
            except NotImplementedError:
                return self.lower_setitem(
                    inst.target, inst.index_var, inst.value, signature
                )
            else:
                target = self.loadvar(inst.target.name)
                value = self.loadvar(inst.value.name)
                valuety = self.typeof(inst.value.name)
                value = self.context.cast(
                    self.builder, value, valuety, signature.args[2]
                )
                return impl(self.builder, (target, inst.index, value))

        elif isinstance(inst, ir.Print):
            self.lower_print(inst)

        elif isinstance(inst, ir.SetItem):
            signature = self.fndesc.calltypes[inst]
            assert signature is not None
            return self.lower_setitem(
                inst.target, inst.index, inst.value, signature
            )

        elif isinstance(inst, ir.StoreMap):
            signature = self.fndesc.calltypes[inst]
            assert signature is not None
            return self.lower_setitem(inst.dct, inst.key, inst.value, signature)

        elif isinstance(inst, ir.DelItem):
            target = self.loadvar(inst.target.name)
            index = self.loadvar(inst.index.name)

            targetty = self.typeof(inst.target.name)
            indexty = self.typeof(inst.index.name)

            signature = self.fndesc.calltypes[inst]
            assert signature is not None

            op = operator.delitem
            fnop = self.context.typing_context.resolve_value_type(op)
            callsig = fnop.get_call_type(
                self.context.typing_context,
                signature.args,
                {},
            )
            impl = self.context.get_function(fnop, callsig)

            assert targetty == signature.args[0]
            index = self.context.cast(
                self.builder, index, indexty, signature.args[1]
            )

            return impl(self.builder, (target, index))

        elif isinstance(inst, ir.Del):
            self.delvar(inst.value)

        elif isinstance(inst, ir.SetAttr):
            target = self.loadvar(inst.target.name)
            value = self.loadvar(inst.value.name)
            signature = self.fndesc.calltypes[inst]

            targetty = self.typeof(inst.target.name)
            valuety = self.typeof(inst.value.name)
            assert signature is not None
            assert signature.args[0] == targetty
            impl = self.context.get_setattr(inst.attr, signature)

            # Convert argument to match
            value = self.context.cast(
                self.builder, value, valuety, signature.args[1]
            )

            return impl(self.builder, (target, value))

        elif isinstance(inst, ir.DynamicRaise):
            self.lower_dynamic_raise(inst)

        elif isinstance(inst, ir.DynamicTryRaise):
            self.lower_try_dynamic_raise(inst)

        elif isinstance(inst, ir.StaticRaise):
            self.lower_static_raise(inst)

        elif isinstance(inst, ir.StaticTryRaise):
            self.lower_static_try_raise(inst)

        else:
            raise NotImplementedError(type(inst))

    def lower_setitem(self, target_var, index_var, value_var, signature):
        target = self.loadvar(target_var.name)
        value = self.loadvar(value_var.name)
        index = self.loadvar(index_var.name)

        targetty = self.typeof(target_var.name)
        valuety = self.typeof(value_var.name)
        indexty = self.typeof(index_var.name)

        op = operator.setitem
        fnop = self.context.typing_context.resolve_value_type(op)
        callsig = fnop.get_call_type(
            self.context.typing_context,
            signature.args,
            {},
        )
        impl = self.context.get_function(fnop, callsig)

        # Convert argument to match
        if isinstance(targetty, types.Optional):
            target = self.context.cast(
                self.builder, target, targetty, targetty.type
            )
        else:
            ul = types.unliteral
            assert ul(targetty) == ul(signature.args[0])

        index = self.context.cast(
            self.builder, index, indexty, signature.args[1]
        )
        value = self.context.cast(
            self.builder, value, valuety, signature.args[2]
        )

        return impl(self.builder, (target, index, value))

    def lower_try_dynamic_raise(self, inst):
        # Numba is a bit limited in what it can do with exceptions in a try
        # block. Thus, it is safe to use the same code as the static try raise.
        self.lower_static_try_raise(inst)

    def lower_dynamic_raise(self, inst):
        exc_args = inst.exc_args
        args = []
        nb_types = []
        for exc_arg in exc_args:
            if isinstance(exc_arg, ir.Var):
                # dynamic values
                typ = self.typeof(exc_arg.name)
                val = self.loadvar(exc_arg.name)
                self.incref(typ, val)
            else:
                typ = None
                val = exc_arg
            nb_types.append(typ)
            args.append(val)

        self.return_dynamic_exception(
            inst.exc_class, tuple(args), tuple(nb_types), loc=self.loc
        )

    def lower_static_raise(self, inst):
        if inst.exc_class is None:
            # Reraise
            self.return_exception(None, loc=self.loc)
        else:
            self.return_exception(inst.exc_class, inst.exc_args, loc=self.loc)

    def lower_static_try_raise(self, inst):
        if inst.exc_class is None:
            # Reraise
            self.set_exception(None, loc=self.loc)
        else:
            self.set_exception(inst.exc_class, inst.exc_args, loc=self.loc)

    def lower_assign(self, ty, inst):
        value = inst.value
        # In nopython mode, closure vars are frozen like globals
        if isinstance(value, (ir.Const, ir.Global, ir.FreeVar)):
            res = self.context.get_constant_generic(
                self.builder, ty, value.value
            )
            self.incref(ty, res)
            return res

        elif isinstance(value, ir.Expr):
            return self.lower_expr(ty, value)

        elif isinstance(value, ir.Var):
            val = self.loadvar(value.name)
            oty = self.typeof(value.name)
            res = self.context.cast(self.builder, val, oty, ty)
            self.incref(ty, res)
            return res

        elif isinstance(value, ir.Arg):
            # Suspend debug info else all the arg repacking ends up being
            # associated with some line or other and it's actually just a detail
            # of Numba's CC.
            with debuginfo.suspend_emission(self.builder):
                # Cast from the argument type to the local variable type
                # (note the "arg.FOO" convention as used in typeinfer)
                argty = self.typeof("arg." + value.name)
                if isinstance(argty, types.Omitted):
                    pyval = argty.value
                    tyctx = self.context.typing_context
                    valty = tyctx.resolve_value_type_prefer_literal(pyval)
                    # use the type of the constant value
                    const = self.context.get_constant_generic(
                        self.builder,
                        valty,
                        pyval,
                    )
                    # cast it to the variable type
                    res = self.context.cast(self.builder, const, valty, ty)
                else:
                    val = self.fnargs[value.index]
                    res = self.context.cast(self.builder, val, argty, ty)
                self.incref(ty, res)
                return res

        elif isinstance(value, ir.Yield):
            res = self.lower_yield(ty, value)
            self.incref(ty, res)
            return res

        raise NotImplementedError(type(value), value)

    def lower_yield(self, retty, inst):
        yp = self.generator_info.yield_points[inst.index]
        assert yp.inst is inst
        y = generators.LowerYield(self, yp, yp.live_vars)
        y.lower_yield_suspend()
        # Yield to caller
        val = self.loadvar(inst.value.name)
        typ = self.typeof(inst.value.name)
        actual_rettyp = self.gentype.yield_type

        # cast the local val to the type yielded
        yret = self.context.cast(self.builder, val, typ, actual_rettyp)

        # get the return repr of yielded value
        retval = self.context.get_return_value(
            self.builder,
            actual_rettyp,
            yret,
        )

        # return
        self.call_conv.return_value(self.builder, retval)

        # Resumption point
        y.lower_yield_resume()
        # None is returned by the yield expression
        return self.context.get_constant_generic(self.builder, retty, None)

    def lower_binop(self, resty, expr, op):
        # if op in utils.OPERATORS_TO_BUILTINS:
        # map operator.the_op => the corresponding types.Function()
        # TODO: is this looks dodgy ...
        op = self.context.typing_context.resolve_value_type(op)

        lhs = expr.lhs
        rhs = expr.rhs
        static_lhs = expr.static_lhs
        static_rhs = expr.static_rhs
        lty = self.typeof(lhs.name)
        rty = self.typeof(rhs.name)
        lhs = self.loadvar(lhs.name)
        rhs = self.loadvar(rhs.name)

        # Convert argument to match
        signature = self.fndesc.calltypes[expr]
        lhs = self.context.cast(self.builder, lhs, lty, signature.args[0])
        rhs = self.context.cast(self.builder, rhs, rty, signature.args[1])

        def cast_result(res):
            return self.context.cast(
                self.builder, res, signature.return_type, resty
            )

        # First try with static operands, if known
        def try_static_impl(tys, args):
            if any(a is ir.UNDEFINED for a in args):
                return None
            try:
                if isinstance(op, types.Function):
                    static_sig = op.get_call_type(
                        self.context.typing_context, tys, {}
                    )
                else:
                    static_sig = typing.signature(signature.return_type, *tys)
            except TypingError:
                return None
            try:
                static_impl = self.context.get_function(op, static_sig)
                return static_impl(self.builder, args)
            except NotImplementedError:
                return None

        res = try_static_impl(
            (_lit_or_omitted(static_lhs), _lit_or_omitted(static_rhs)),
            (static_lhs, static_rhs),
        )
        if res is not None:
            return cast_result(res)

        res = try_static_impl(
            (_lit_or_omitted(static_lhs), rty),
            (static_lhs, rhs),
        )
        if res is not None:
            return cast_result(res)

        res = try_static_impl(
            (lty, _lit_or_omitted(static_rhs)),
            (lhs, static_rhs),
        )
        if res is not None:
            return cast_result(res)

        # Normal implementation for generic arguments

        sig = op.get_call_type(self.context.typing_context, signature.args, {})
        impl = self.context.get_function(op, sig)
        res = impl(self.builder, (lhs, rhs))
        return cast_result(res)

    def lower_getitem(self, resty, expr, value, index, signature):
        baseval = self.loadvar(value.name)
        indexval = self.loadvar(index.name)
        # Get implementation of getitem
        op = operator.getitem
        fnop = self.context.typing_context.resolve_value_type(op)
        callsig = fnop.get_call_type(
            self.context.typing_context,
            signature.args,
            {},
        )
        impl = self.context.get_function(fnop, callsig)

        argvals = (baseval, indexval)
        argtyps = (self.typeof(value.name), self.typeof(index.name))
        castvals = [
            self.context.cast(self.builder, av, at, ft)
            for av, at, ft in zip(argvals, argtyps, signature.args)
        ]
        res = impl(self.builder, castvals)
        return self.context.cast(
            self.builder, res, signature.return_type, resty
        )

    def _cast_var(self, var, ty):
        """
        Cast a Numba IR variable to the given Numba type, returning a
        low-level value.
        """
        if isinstance(var, _VarArgItem):
            varty = self.typeof(var.vararg.name)[var.index]
            val = self.builder.extract_value(
                self.loadvar(var.vararg.name), var.index
            )
        else:
            varty = self.typeof(var.name)
            val = self.loadvar(var.name)
        return self.context.cast(self.builder, val, varty, ty)

    def fold_call_args(self, fnty, signature, pos_args, vararg, kw_args):
        if vararg:
            # Inject *args from function call
            # The lowering will be done in _cast_var() above.
            tp_vararg = self.typeof(vararg.name)
            assert isinstance(tp_vararg, types.BaseTuple)
            pos_args = pos_args + [
                _VarArgItem(vararg, i) for i in range(len(tp_vararg))
            ]

        # Fold keyword arguments and resolve default argument values
        pysig = signature.pysig
        if pysig is None:
            if kw_args:
                raise NotImplementedError(
                    "unsupported keyword arguments when calling %s" % (fnty,)
                )
            argvals = [
                self._cast_var(var, sigty)
                for var, sigty in zip(pos_args, signature.args)
            ]
        else:

            def normal_handler(index, param, var):
                return self._cast_var(var, signature.args[index])

            def default_handler(index, param, default):
                return self.context.get_constant_generic(
                    self.builder, signature.args[index], default
                )

            def stararg_handler(index, param, vars):
                stararg_ty = signature.args[index]
                assert isinstance(stararg_ty, types.BaseTuple), stararg_ty
                values = [
                    self._cast_var(var, sigty)
                    for var, sigty in zip(vars, stararg_ty)
                ]
                return cgutils.make_anonymous_struct(self.builder, values)

            argvals = typing.fold_arguments(
                pysig,
                pos_args,
                dict(kw_args),
                normal_handler,
                default_handler,
                stararg_handler,
            )
        return argvals

    def lower_print(self, inst):
        """
        Lower a ir.Print()
        """
        # We handle this, as far as possible, as a normal call to built-in
        # print().  This will make it easy to undo the special ir.Print
        # rewrite when it becomes unnecessary (e.g. when we have native
        # strings).
        sig = self.fndesc.calltypes[inst]
        assert sig.return_type == types.none
        fnty = self.context.typing_context.resolve_value_type(print)

        # Fix the call signature to inject any constant-inferred
        # string argument
        pos_tys = list(sig.args)
        pos_args = list(inst.args)
        for i in range(len(pos_args)):
            if i in inst.consts:
                pyval = inst.consts[i]
                if isinstance(pyval, str):
                    pos_tys[i] = types.literal(pyval)

        fixed_sig = typing.signature(sig.return_type, *pos_tys)
        fixed_sig = fixed_sig.replace(pysig=sig.pysig)

        argvals = self.fold_call_args(fnty, sig, pos_args, inst.vararg, {})
        impl = self.context.get_function(print, fixed_sig)
        impl(self.builder, argvals)

    def lower_call(self, resty, expr):
        signature = self.fndesc.calltypes[expr]
        self.debug_print("# lower_call: expr = {0}".format(expr))
        if isinstance(signature.return_type, types.Phantom):
            return self.context.get_dummy_value()

        fnty = self.typeof(expr.func.name)

        if isinstance(fnty, types.ObjModeDispatcher):
            res = self._lower_call_ObjModeDispatcher(fnty, expr, signature)

        elif isinstance(fnty, types.ExternalFunction):
            res = self._lower_call_ExternalFunction(fnty, expr, signature)

        elif isinstance(fnty, types.ExternalFunctionPointer):
            res = self._lower_call_ExternalFunctionPointer(
                fnty, expr, signature
            )

        elif isinstance(fnty, types.RecursiveCall):
            res = self._lower_call_RecursiveCall(fnty, expr, signature)

        else:
            res = self._lower_call_normal(fnty, expr, signature)

        # If lowering the call returned None, interpret that as returning dummy
        # value if the return type of the function is void, otherwise there is
        # a problem
        if res is None:
            if signature.return_type == types.void:
                res = self.context.get_dummy_value()
            else:
                raise LoweringError(
                    msg="non-void function returns None from implementation",
                    loc=self.loc,
                )

        return self.context.cast(
            self.builder, res, signature.return_type, resty
        )

    def _lower_call_ObjModeDispatcher(self, fnty, expr, signature):
        from numba.cuda.core.pythonapi import ObjModeUtils

        self.init_pyapi()
        # Acquire the GIL
        gil_state = self.pyapi.gil_ensure()
        # Fix types
        argnames = [a.name for a in expr.args]
        argtypes = [self.typeof(a) for a in argnames]
        argvalues = [self.loadvar(a) for a in argnames]
        for v, ty in zip(argvalues, argtypes):
            # Because .from_native_value steal the reference
            self.incref(ty, v)

        argobjs = [
            self.pyapi.from_native_value(atyp, aval, self.env_manager)
            for atyp, aval in zip(argtypes, argvalues)
        ]

        # Load objmode dispatcher
        callee = ObjModeUtils(self.pyapi).load_dispatcher(fnty, argtypes)
        # Make Call
        ret_obj = self.pyapi.call_function_objargs(callee, argobjs)
        has_exception = cgutils.is_null(self.builder, ret_obj)
        with self.builder.if_else(has_exception) as (then, orelse):
            # Handles exception
            # This branch must exit the function
            with then:
                # Clean arg
                for obj in argobjs:
                    self.pyapi.decref(obj)

                # Release the GIL
                self.pyapi.gil_release(gil_state)

                # Return and signal exception
                self.call_conv.return_exc(self.builder)

            # Handles normal return
            with orelse:
                # Fix output value
                native = self.pyapi.to_native_value(
                    fnty.dispatcher.output_types,
                    ret_obj,
                )
                output = native.value

                # Release objs
                self.pyapi.decref(ret_obj)
                for obj in argobjs:
                    self.pyapi.decref(obj)

                # cleanup output
                if callable(native.cleanup):
                    native.cleanup()

                # Release the GIL
                self.pyapi.gil_release(gil_state)

                # Error during unboxing
                with self.builder.if_then(native.is_error):
                    self.call_conv.return_exc(self.builder)

                return output

    def _lower_call_ExternalFunction(self, fnty, expr, signature):
        # Handle a named external function
        self.debug_print("# external function")
        argvals = self.fold_call_args(
            fnty,
            signature,
            expr.args,
            expr.vararg,
            expr.kws,
        )
        fndesc = funcdesc.ExternalFunctionDescriptor(
            fnty.symbol, fnty.sig.return_type, fnty.sig.args
        )
        func = self.context.declare_external_function(
            self.builder.module, fndesc
        )
        return self.context.call_external_function(
            self.builder,
            func,
            fndesc.argtypes,
            argvals,
        )

    def _lower_call_ExternalFunctionPointer(self, fnty, expr, signature):
        # Handle a C function pointer
        self.debug_print("# calling external function pointer")
        argvals = self.fold_call_args(
            fnty,
            signature,
            expr.args,
            expr.vararg,
            expr.kws,
        )
        pointer = self.loadvar(expr.func.name)
        # If the external function pointer uses libpython
        if fnty.requires_gil:
            self.init_pyapi()
            # Acquire the GIL
            gil_state = self.pyapi.gil_ensure()
            # Make PyObjects
            newargvals = []
            pyvals = []
            for exptyp, gottyp, aval in zip(
                fnty.sig.args, signature.args, argvals
            ):
                # Adjust argument values to pyobjects
                if exptyp == types.ffi_forced_object:
                    self.incref(gottyp, aval)
                    obj = self.pyapi.from_native_value(
                        gottyp,
                        aval,
                        self.env_manager,
                    )
                    newargvals.append(obj)
                    pyvals.append(obj)
                else:
                    newargvals.append(aval)

            # Call external function
            res = self.context.call_function_pointer(
                self.builder,
                pointer,
                newargvals,
                fnty.cconv,
            )
            # Release PyObjects
            for obj in pyvals:
                self.pyapi.decref(obj)

            # Release the GIL
            self.pyapi.gil_release(gil_state)
        # If the external function pointer does NOT use libpython
        else:
            res = self.context.call_function_pointer(
                self.builder,
                pointer,
                argvals,
                fnty.cconv,
            )
        return res

    def _lower_call_RecursiveCall(self, fnty, expr, signature):
        # Recursive call
        argvals = self.fold_call_args(
            fnty,
            signature,
            expr.args,
            expr.vararg,
            expr.kws,
        )
        rec_ov = fnty.get_overloads(signature.args)
        mangler = self.context.mangler or default_mangler
        abi_tags = self.fndesc.abi_tags
        mangled_name = mangler(
            rec_ov.qualname, signature.args, abi_tags=abi_tags, uid=rec_ov.uid
        )
        # special case self recursion
        if self.builder.function.name.startswith(mangled_name):
            res = self.context.call_internal(
                self.builder,
                self.fndesc,
                signature,
                argvals,
            )
        else:
            res = self.context.call_unresolved(
                self.builder,
                mangled_name,
                signature,
                argvals,
            )
        return res

    def _lower_call_normal(self, fnty, expr, signature):
        # Normal function resolution
        self.debug_print("# calling normal function: {0}".format(fnty))
        self.debug_print("# signature: {0}".format(signature))
        if isinstance(fnty, types.ObjModeDispatcher):
            argvals = expr.func.args
        else:
            argvals = self.fold_call_args(
                fnty,
                signature,
                expr.args,
                expr.vararg,
                expr.kws,
            )
        tname = expr.target
        if tname is not None:
            from numba.cuda.descriptor import cuda_target

            hw_ctx = cuda_target.target_context
            impl = hw_ctx.get_function(fnty, signature)
        else:
            impl = self.context.get_function(fnty, signature)
        if signature.recvr:
            # The "self" object is passed as the function object
            # for bounded function
            the_self = self.loadvar(expr.func.name)
            # Prepend the self reference
            argvals = [the_self] + list(argvals)

        res = impl(self.builder, argvals, self.loc)
        return res

    def lower_expr(self, resty, expr):
        if expr.op == "binop":
            return self.lower_binop(resty, expr, expr.fn)
        elif expr.op == "inplace_binop":
            lty = self.typeof(expr.lhs.name)
            if lty.mutable:
                return self.lower_binop(resty, expr, expr.fn)
            else:
                # inplace operators on non-mutable types reuse the same
                # definition as the corresponding copying operators.)
                return self.lower_binop(resty, expr, expr.immutable_fn)
        elif expr.op == "unary":
            val = self.loadvar(expr.value.name)
            typ = self.typeof(expr.value.name)
            func_ty = self.context.typing_context.resolve_value_type(expr.fn)
            # Get function
            signature = self.fndesc.calltypes[expr]
            impl = self.context.get_function(func_ty, signature)
            # Convert argument to match
            val = self.context.cast(self.builder, val, typ, signature.args[0])
            res = impl(self.builder, [val])
            res = self.context.cast(
                self.builder, res, signature.return_type, resty
            )
            return res

        elif expr.op == "call":
            res = self.lower_call(resty, expr)
            return res

        elif expr.op == "pair_first":
            val = self.loadvar(expr.value.name)
            ty = self.typeof(expr.value.name)
            res = self.context.pair_first(self.builder, val, ty)
            self.incref(resty, res)
            return res

        elif expr.op == "pair_second":
            val = self.loadvar(expr.value.name)
            ty = self.typeof(expr.value.name)
            res = self.context.pair_second(self.builder, val, ty)
            self.incref(resty, res)
            return res

        elif expr.op in ("getiter", "iternext"):
            val = self.loadvar(expr.value.name)
            ty = self.typeof(expr.value.name)
            signature = self.fndesc.calltypes[expr]
            impl = self.context.get_function(expr.op, signature)
            [fty] = signature.args
            castval = self.context.cast(self.builder, val, ty, fty)
            res = impl(self.builder, (castval,))
            res = self.context.cast(
                self.builder, res, signature.return_type, resty
            )
            return res

        elif expr.op == "exhaust_iter":
            val = self.loadvar(expr.value.name)
            ty = self.typeof(expr.value.name)
            # Unpack optional
            if isinstance(ty, types.Optional):
                val = self.context.cast(self.builder, val, ty, ty.type)
                ty = ty.type

            # If we have a tuple, we needn't do anything
            # (and we can't iterate over the heterogeneous ones).
            if isinstance(ty, types.BaseTuple):
                assert ty == resty
                self.incref(ty, val)
                return val

            itemty = ty.iterator_type.yield_type
            tup = self.context.get_constant_undef(resty)
            pairty = types.Pair(itemty, types.boolean)
            getiter_sig = typing.signature(ty.iterator_type, ty)
            getiter_impl = self.context.get_function("getiter", getiter_sig)
            iternext_sig = typing.signature(pairty, ty.iterator_type)
            iternext_impl = self.context.get_function("iternext", iternext_sig)
            iterobj = getiter_impl(self.builder, (val,))
            # We call iternext() as many times as desired (`expr.count`).
            for i in range(expr.count):
                pair = iternext_impl(self.builder, (iterobj,))
                is_valid = self.context.pair_second(self.builder, pair, pairty)
                with cgutils.if_unlikely(
                    self.builder, self.builder.not_(is_valid)
                ):
                    self.return_exception(ValueError, loc=self.loc)
                item = self.context.pair_first(self.builder, pair, pairty)
                tup = self.builder.insert_value(tup, item, i)

            # Call iternext() once more to check that the iterator
            # is exhausted.
            pair = iternext_impl(self.builder, (iterobj,))
            is_valid = self.context.pair_second(self.builder, pair, pairty)
            with cgutils.if_unlikely(self.builder, is_valid):
                self.return_exception(ValueError, loc=self.loc)

            self.decref(ty.iterator_type, iterobj)
            return tup

        elif expr.op == "getattr":
            val = self.loadvar(expr.value.name)
            ty = self.typeof(expr.value.name)

            if isinstance(resty, types.BoundFunction):
                # if we are getting out a method, assume we have typed this
                # properly and just build a bound function object
                casted = self.context.cast(self.builder, val, ty, resty.this)
                res = self.context.get_bound_function(
                    self.builder, casted, resty.this
                )
                self.incref(resty, res)
                return res
            else:
                impl = self.context.get_getattr(ty, expr.attr)
                attrty = self.context.typing_context.resolve_getattr(
                    ty, expr.attr
                )

                if impl is None:
                    # ignore the attribute
                    return self.context.get_dummy_value()
                else:
                    res = impl(self.context, self.builder, ty, val, expr.attr)

                # Cast the attribute type to the expected output type
                res = self.context.cast(self.builder, res, attrty, resty)
                return res

        elif expr.op == "static_getitem":
            signature = typing.signature(
                resty,
                self.typeof(expr.value.name),
                _lit_or_omitted(expr.index),
            )
            try:
                # Both get_function() and the returned implementation can
                # raise NotImplementedError if the types aren't supported
                impl = self.context.get_function("static_getitem", signature)
                return impl(
                    self.builder, (self.loadvar(expr.value.name), expr.index)
                )
            except NotImplementedError:
                if expr.index_var is None:
                    raise
                # Fall back on the generic getitem() implementation
                # for this type.
                signature = self.fndesc.calltypes[expr]
                return self.lower_getitem(
                    resty, expr, expr.value, expr.index_var, signature
                )
        elif expr.op == "typed_getitem":
            signature = typing.signature(
                resty,
                self.typeof(expr.value.name),
                self.typeof(expr.index.name),
            )
            impl = self.context.get_function("typed_getitem", signature)
            return impl(
                self.builder,
                (self.loadvar(expr.value.name), self.loadvar(expr.index.name)),
            )
        elif expr.op == "getitem":
            signature = self.fndesc.calltypes[expr]
            return self.lower_getitem(
                resty, expr, expr.value, expr.index, signature
            )

        elif expr.op == "build_tuple":
            itemvals = [self.loadvar(i.name) for i in expr.items]
            itemtys = [self.typeof(i.name) for i in expr.items]
            castvals = [
                self.context.cast(self.builder, val, fromty, toty)
                for val, toty, fromty in zip(itemvals, resty, itemtys)
            ]
            tup = self.context.make_tuple(self.builder, resty, castvals)
            self.incref(resty, tup)
            return tup

        elif expr.op == "build_list":
            itemvals = [self.loadvar(i.name) for i in expr.items]
            itemtys = [self.typeof(i.name) for i in expr.items]
            if isinstance(resty, types.LiteralList):
                castvals = [
                    self.context.cast(self.builder, val, fromty, toty)
                    for val, toty, fromty in zip(itemvals, resty.types, itemtys)
                ]
                tup = self.context.make_tuple(
                    self.builder, types.Tuple(resty.types), castvals
                )
                self.incref(resty, tup)
                return tup
            else:
                castvals = [
                    self.context.cast(self.builder, val, fromty, resty.dtype)
                    for val, fromty in zip(itemvals, itemtys)
                ]
                return self.context.build_list(self.builder, resty, castvals)

        elif expr.op == "build_set":
            # Insert in reverse order, as Python does
            items = expr.items[::-1]
            itemvals = [self.loadvar(i.name) for i in items]
            itemtys = [self.typeof(i.name) for i in items]
            castvals = [
                self.context.cast(self.builder, val, fromty, resty.dtype)
                for val, fromty in zip(itemvals, itemtys)
            ]
            return self.context.build_set(self.builder, resty, castvals)

        elif expr.op == "build_map":
            items = expr.items
            keys, values = [], []
            key_types, value_types = [], []
            for k, v in items:
                key = self.loadvar(k.name)
                keytype = self.typeof(k.name)
                val = self.loadvar(v.name)
                valtype = self.typeof(v.name)
                keys.append(key)
                values.append(val)
                key_types.append(keytype)
                value_types.append(valtype)
            return self.context.build_map(
                self.builder,
                resty,
                list(zip(key_types, value_types)),
                list(zip(keys, values)),
            )

        elif expr.op == "cast":
            val = self.loadvar(expr.value.name)
            ty = self.typeof(expr.value.name)
            castval = self.context.cast(self.builder, val, ty, resty)
            self.incref(resty, castval)
            return castval

        elif expr.op == "phi":
            raise LoweringError("PHI not stripped")

        elif expr.op == "null":
            return self.context.get_constant_null(resty)

        elif expr.op == "undef":
            # Numba does not raise an UnboundLocalError for undefined variables.
            # The variable is set to zero.
            return self.context.get_constant_null(resty)

        elif expr.op in self.context.special_ops:
            res = self.context.special_ops[expr.op](self, expr)
            return res

        raise NotImplementedError(expr)

    def _alloca_var(self, name, fetype):
        """
        Ensure the given variable has an allocated stack slot (if needed).
        """
        if name in self.varmap:
            # quit early
            return

        # If the name is used in multiple blocks or lowering with debuginfo...
        if (
            name not in self._singly_assigned_vars
        ) or self._disable_sroa_like_opt:
            # If not already defined, allocate it
            ptr = self.alloca(name, fetype)
            # Remember the pointer
            self.varmap[name] = ptr

    def getvar(self, name):
        """
        Get a pointer to the given variable's slot.
        """
        if not self._disable_sroa_like_opt:
            assert name not in self._blk_local_varmap
            assert name not in self._singly_assigned_vars
        if name not in self.varmap:
            # Allocate undefined variable as needed.
            # NOTE: Py3.12 use of LOAD_FAST_AND_CLEAR will allow variable be
            # referenced before it is defined.
            self._alloca_var(name, self.typeof(name))
        return self.varmap[name]

    def loadvar(self, name):
        """
        Load the given variable's value.
        """
        if name in self._blk_local_varmap and not self._disable_sroa_like_opt:
            return self._blk_local_varmap[name]
        ptr = self.getvar(name)

        # Don't associate debuginfo with the load for a function arg else it
        # creates instructions ahead of the first source line of the
        # function which then causes problems with breaking on the function
        # symbol (it hits the symbol, not the first line).
        if name in self.func_ir.arg_names:
            with debuginfo.suspend_emission(self.builder):
                return self.builder.load(ptr)
        else:
            return self.builder.load(ptr)

    def storevar(self, value, name, argidx=None):
        """
        Store the value into the given variable.
        """
        fetype = self.typeof(name)
        # Define if not already
        self._alloca_var(name, fetype)

        # Store variable
        if (
            name in self._singly_assigned_vars
            and not self._disable_sroa_like_opt
        ):
            self._blk_local_varmap[name] = value
        else:
            if argidx is None:
                # Clean up existing value stored in the variable, not needed
                # if it's an arg
                old = self.loadvar(name)
                self.decref(fetype, old)

            # stack stored variable
            ptr = self.getvar(name)
            if value.type != ptr.type.pointee:
                msg = (
                    "Storing {value.type} to ptr of {ptr.type.pointee} "
                    "('{name}'). FE type {fetype}"
                ).format(value=value, ptr=ptr, fetype=fetype, name=name)
                raise AssertionError(msg)

            # If this store is associated with an argument to the function (i.e.
            # store following reassemble from CC splatting structs as many args
            # to the function) then mark this variable as such.
            if argidx is not None:
                with debuginfo.suspend_emission(self.builder):
                    self.builder.store(value, ptr)
                loc = self.defn_loc  # the line with `def <func>`
                lltype = self.context.get_value_type(fetype)
                sizeof = self.context.get_abi_sizeof(lltype)
                datamodel = self.context.data_model_manager[fetype]
                self.debuginfo.mark_variable(
                    self.builder,
                    ptr,
                    name=name,
                    lltype=lltype,
                    size=sizeof,
                    line=loc.line,
                    datamodel=datamodel,
                    argidx=argidx,
                )
            else:
                self.builder.store(value, ptr)

    def delvar(self, name):
        """
        Delete the given variable.
        """
        fetype = self.typeof(name)

        # Out-of-order
        if (
            name not in self._blk_local_varmap
            and not self._disable_sroa_like_opt
        ):
            if name in self._singly_assigned_vars:
                self._singly_assigned_vars.discard(name)

        # Define if not already (may happen if the variable is deleted
        # at the beginning of a loop, but only set later in the loop)
        self._alloca_var(name, fetype)

        if name in self._blk_local_varmap and not self._disable_sroa_like_opt:
            llval = self._blk_local_varmap[name]
            self.decref(fetype, llval)
        else:
            ptr = self.getvar(name)
            self.decref(fetype, self.builder.load(ptr))
            # Zero-fill variable to avoid double frees on subsequent dels
            self.builder.store(llvm_ir.Constant(ptr.type.pointee, None), ptr)

    def alloca(self, name, type):
        lltype = self.context.get_value_type(type)
        datamodel = self.context.data_model_manager[type]
        return self.alloca_lltype(name, lltype, datamodel=datamodel)

    def alloca_lltype(self, name, lltype, datamodel=None):
        # Is user variable?
        is_uservar = not name.startswith("$")
        # Allocate space for variable
        aptr = cgutils.alloca_once(self.builder, lltype, name=name, zfill=False)

        # Emit debug info for user variable
        if is_uservar:
            # Don't associate debuginfo with the alloca for a function arg, this
            # is handled by the first store to the alloca so that repacking the
            # splatted args from the CC is dealt with.
            if name not in self.func_ir.arg_names:
                sizeof = self.context.get_abi_sizeof(lltype)
                self.debuginfo.mark_variable(
                    self.builder,
                    aptr,
                    name=name,
                    lltype=lltype,
                    size=sizeof,
                    line=self.loc.line,
                    datamodel=datamodel,
                )
        return aptr

    def incref(self, typ, val):
        if not self.context.enable_nrt:
            return

        self.context.nrt.incref(self.builder, typ, val)

    def decref(self, typ, val):
        if not self.context.enable_nrt:
            return

        # do not associate decref with "use", it creates "jumpy" line info as
        # the decrefs are usually where the ir.Del nodes are, which is at the
        # end of the block.
        with debuginfo.suspend_emission(self.builder):
            self.context.nrt.decref(self.builder, typ, val)


class CUDALower(Lower):
    def storevar(self, value, name, argidx=None):
        """
        Store the value into the given variable.
        """
        # Handle polymorphic variables with CUDA_DEBUG_POLY enabled
        if config.CUDA_DEBUG_POLY:
            src_name = name.split(".")[0]
            if src_name in self.poly_var_typ_map:
                # Ensure allocation happens first (if needed)
                fetype = self.typeof(name)
                self._alloca_var(name, fetype)
                # Discriminant and data are located in the same union
                ptr = self.poly_var_loc_map[src_name]
                # Firstly write discriminant to the beginning of union as i8
                dtype = types.UnionType(self.poly_var_typ_map[src_name])
                # Compute discriminant = index of type in sorted union
                if isinstance(fetype, types.Literal):
                    lookup_type = fetype.literal_type
                else:
                    lookup_type = fetype
                discriminant_val = list(dtype.types).index(lookup_type)
                # Bitcast union pointer directly to i8* and write
                # discriminant at offset 0
                discriminant_ptr = self.builder.bitcast(
                    ptr, llvm_ir.PointerType(llvm_ir.IntType(8))
                )
                discriminant_i8 = llvm_ir.Constant(
                    llvm_ir.IntType(8), discriminant_val
                )
                self.builder.store(discriminant_i8, discriminant_ptr)
                # Secondly write data at offset = sizeof(fetype) in bytes
                lltype = self.context.get_value_type(fetype)
                sizeof_bytes = self.context.get_abi_sizeof(lltype)
                # Bitcast to i8* and use byte-level GEP
                byte_ptr = self.builder.bitcast(
                    ptr, llvm_ir.PointerType(llvm_ir.IntType(8))
                )
                data_byte_ptr = self.builder.gep(
                    byte_ptr,
                    [llvm_ir.Constant(llvm_ir.IntType(64), sizeof_bytes)],
                )
                # Cast to the correct type pointer
                castptr = self.builder.bitcast(
                    data_byte_ptr, llvm_ir.PointerType(lltype)
                )
                self.builder.store(value, castptr)
                return

        # For non-polymorphic variables, use parent implementation
        super().storevar(value, name, argidx)

        # Emit llvm.dbg.value instead of llvm.dbg.declare for local scalar
        # variables immediately after a store instruction.
        if (
            self.context.enable_debuginfo
            # Conditions used to elide stores in parent method
            and self.store_var_needed(name)
        ):
            fetype = self.typeof(name)
            lltype = self.context.get_value_type(fetype)
            int_type = (llvm_ir.IntType,)
            real_type = llvm_ir.FloatType, llvm_ir.DoubleType
            if isinstance(lltype, int_type + real_type):
                sizeof = self.context.get_abi_sizeof(lltype)
                datamodel = self.context.data_model_manager[fetype]
                line = self.loc.line if argidx is None else self.defn_loc.line
                if not name.startswith("$"):
                    # Emit debug value for user variable
                    src_name = name.split(".")[0]
                    if src_name not in self.poly_var_typ_map:
                        # Insert the llvm.dbg.value intrinsic call
                        self.debuginfo.update_variable(
                            self.builder,
                            value,
                            src_name,
                            lltype,
                            sizeof,
                            line,
                            datamodel,
                            argidx,
                        )
                elif isinstance(value, llvm_ir.LoadInstr):
                    # Emit debug value for user variable that falls out of the
                    # coverage of dbg.value range per basic block
                    ld_name = value.operands[0].name
                    if not ld_name.startswith(("$", ".")):
                        src_name = ld_name.split(".")[0]
                        if (
                            src_name not in self.poly_var_typ_map
                            # Not yet covered by the dbg.value range
                            and src_name not in self.dbg_val_names
                        ):
                            for index, item in enumerate(self.fnargs):
                                if item.name == src_name:
                                    argidx = index + 1
                                    break
                            # Insert the llvm.dbg.value intrinsic call
                            self.debuginfo.update_variable(
                                self.builder,
                                value,
                                src_name,
                                lltype,
                                sizeof,
                                line,
                                datamodel,
                                argidx,
                            )

    def pre_block(self, block):
        super().pre_block(block)

        # dbg.value range covered names
        self.dbg_val_names = set()

        if self.context.enable_debuginfo and self._disable_sroa_like_opt:
            for x in block.find_insts(ir.Assign):
                if x.target.name.startswith("$"):
                    continue
                ssa_name = x.target.name
                src_name = ssa_name.split(".")[0]
                if src_name not in self.dbg_val_names:
                    self.dbg_val_names.add(src_name)

    def pre_lower(self):
        """
        Called before lowering all blocks.
        """
        super().pre_lower()

        self.poly_var_typ_map = {}
        self.poly_var_loc_map = {}
        self.poly_var_set = set()
        self.poly_cleaned = False
        self.lastblk = max(self.blocks.keys())

        # When debug info is enabled, walk through function body and mark
        # variables with polymorphic types.
        if self.context.enable_debuginfo and self._disable_sroa_like_opt:
            poly_map = {}
            # pre-scan all blocks
            for block in self.blocks.values():
                for x in block.find_insts(ir.Assign):
                    if x.target.name.startswith("$"):
                        continue
                    ssa_name = x.target.name
                    src_name = ssa_name.split(".")[0]
                    # Check all the multi-versioned targets
                    if len(x.target.versioned_names) > 0:
                        fetype = self.typeof(ssa_name)
                        if src_name not in poly_map:
                            poly_map[src_name] = set()
                        # deduplicate polymorphic types
                        if isinstance(fetype, types.Literal):
                            fetype = fetype.literal_type
                        poly_map[src_name].add(fetype)
            # Filter out multi-versioned but single typed variables
            self.poly_var_typ_map = {
                k: v for k, v in poly_map.items() if len(v) > 1
            }

    def _alloca_var(self, name, fetype):
        """
        Ensure the given variable has an allocated stack slot (if needed).
        """
        # If the name is not handled yet and a store is needed
        if name not in self.varmap and self.store_var_needed(name):
            src_name = name.split(".")[0]
            if src_name in self.poly_var_typ_map:
                self.poly_var_set.add(name)
                if src_name not in self.poly_var_loc_map:
                    dtype = types.UnionType(self.poly_var_typ_map[src_name])
                    datamodel = self.context.data_model_manager[dtype]
                    # UnionType has sorted set of types, max at last index
                    maxsizetype = dtype.types[-1]
                    if config.CUDA_DEBUG_POLY:
                        # allocate double the max element size to house
                        # [discriminant + data]
                        aggr_type = types.UniTuple(maxsizetype, 2)
                    else:
                        # allocate single element for data only
                        aggr_type = types.UniTuple(maxsizetype, 1)
                    lltype = self.context.get_value_type(aggr_type)
                    ptr = self.alloca_lltype(src_name, lltype, datamodel)
                    # save the location of the union type for polymorphic var
                    self.poly_var_loc_map[src_name] = ptr
                return

        super()._alloca_var(name, fetype)

    def store_var_needed(self, name):
        # Check the conditions used to elide stores in parent class,
        # e.g. in method storevar() and _alloca_var()
        return (
            # used in multiple blocks
            name not in self._singly_assigned_vars
            # lowering with debuginfo
            or self._disable_sroa_like_opt
        )

    def delvar(self, name):
        """
        Delete the given variable.
        """
        if name in self.poly_var_set:
            fetype = self.typeof(name)
            src_name = name.split(".")[0]
            ptr = self.poly_var_loc_map[src_name]
            self.decref(fetype, self.builder.load(ptr))
            if (
                self._cur_ir_block == self.blocks[self.lastblk]
                and not self.poly_cleaned
            ):
                # Zero-fill the debug union for polymorphic only
                # at the last block
                for v in self.poly_var_loc_map.values():
                    self.builder.store(
                        llvm_ir.Constant(v.type.pointee, None), v
                    )
                    self.poly_cleaned = True
            return

        super().delvar(name)

    def getvar(self, name):
        """
        Get a pointer to the given variable's slot.
        """
        if name in self.poly_var_set:
            src_name = name.split(".")[0]
            fetype = self.typeof(name)
            lltype = self.context.get_value_type(fetype)
            ptr = self.poly_var_loc_map[src_name]

            if config.CUDA_DEBUG_POLY:
                # With CUDA_DEBUG_POLY enabled, read value at
                # offset = sizeof(fetype) in bytes
                sizeof_bytes = self.context.get_abi_sizeof(lltype)
                # Bitcast to i8* and use byte-level GEP
                byte_ptr = self.builder.bitcast(
                    ptr, llvm_ir.PointerType(llvm_ir.IntType(8))
                )
                value_byte_ptr = self.builder.gep(
                    byte_ptr,
                    [llvm_ir.Constant(llvm_ir.IntType(64), sizeof_bytes)],
                )
                # Cast to the correct type pointer
                castptr = self.builder.bitcast(
                    value_byte_ptr, llvm_ir.PointerType(lltype)
                )
            else:
                # Otherwise, just bitcast to the correct type
                castptr = self.builder.bitcast(ptr, llvm_ir.PointerType(lltype))
            return castptr
        else:
            return super().getvar(name)


def _lit_or_omitted(value):
    """Returns a Literal instance if the type of value is supported;
    otherwise, return `Omitted(value)`.
    """
    typing_errors = LiteralTypingError
    if HAS_NUMBA:
        from numba.core.errors import (
            LiteralTypingError as CoreLiteralTypingError,
        )

        typing_errors = (LiteralTypingError, CoreLiteralTypingError)
    try:
        return types.literal(value)
    except typing_errors:
        return types.Omitted(value)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import sys
import numpy as np
from ml_dtypes import bfloat16 as mldtypes_bf16
from numba import cuda
from numba.cuda import (
    float32,
    float64,
    int16,
    int32,
    int64,
    uint16,
    uint32,
    uint64,
)
from numba.cuda import config

if not config.ENABLE_CUDASIM:
    from numba.cuda.bf16 import (
        bfloat16,
        habs,
        hadd,
        hsub,
        hmul,
        hadd_rn,
        hsub_rn,
        hmul_rn,
        hdiv,
        hadd_sat,
        hsub_sat,
        hmul_sat,
        hfma,
        hfma_sat,
        hneg,
        hfma_relu,
        # Comparison intrinsics
        heq,
        hne,
        hge,
        hgt,
        hle,
        hlt,
        hmax,
        hmin,
        hmax_nan,
        hmin_nan,
        hisnan,
        hisinf,
        # Conversion intrinsics (NumPy-style names)
        bfloat16_to_int8_rz,
        bfloat16_to_uint8_rz,
        int16_to_bfloat16_rn,
        int16_to_bfloat16_rz,
        int16_to_bfloat16_rd,
        int16_to_bfloat16_ru,
        bfloat16_to_int16_rn,
        bfloat16_to_int16_rz,
        bfloat16_to_int16_rd,
        bfloat16_to_int16_ru,
        uint16_to_bfloat16_rn,
        uint16_to_bfloat16_rz,
        uint16_to_bfloat16_rd,
        uint16_to_bfloat16_ru,
        bfloat16_to_uint16_rn,
        bfloat16_to_uint16_rz,
        bfloat16_to_uint16_rd,
        bfloat16_to_uint16_ru,
        int32_to_bfloat16_rn,
        int32_to_bfloat16_rz,
        int32_to_bfloat16_rd,
        int32_to_bfloat16_ru,
        bfloat16_to_int32_rn,
        bfloat16_to_int32_rz,
        bfloat16_to_int32_rd,
        bfloat16_to_int32_ru,
        uint32_to_bfloat16_rn,
        uint32_to_bfloat16_rz,
        uint32_to_bfloat16_rd,
        uint32_to_bfloat16_ru,
        bfloat16_to_uint32_rn,
        bfloat16_to_uint32_rz,
        bfloat16_to_uint32_rd,
        bfloat16_to_uint32_ru,
        bfloat16_to_int64_rn,
        bfloat16_to_int64_rz,
        bfloat16_to_int64_rd,
        bfloat16_to_int64_ru,
        int64_to_bfloat16_rn,
        int64_to_bfloat16_rz,
        int64_to_bfloat16_rd,
        int64_to_bfloat16_ru,
        bfloat16_to_uint64_rn,
        bfloat16_to_uint64_rz,
        bfloat16_to_uint64_rd,
        bfloat16_to_uint64_ru,
        uint64_to_bfloat16_rn,
        uint64_to_bfloat16_rz,
        uint64_to_bfloat16_rd,
        uint64_to_bfloat16_ru,
        bfloat16_as_int16,
        int16_as_bfloat16,
        bfloat16_as_uint16,
        uint16_as_bfloat16,
        bfloat16_to_float32,
        float32_to_bfloat16,
        float64_to_bfloat16,
        float32_to_bfloat16_rn,
        float32_to_bfloat16_rz,
        float32_to_bfloat16_rd,
        float32_to_bfloat16_ru,
    )

from numba.cuda.testing import CUDATestCase

import math


class TestBfloat16HighLevelBindings(CUDATestCase):
    def skip_unsupported(self):
        if not cuda.is_bfloat16_supported():
            self.skipTest("bfloat16 requires compute capability 8.0+")

    def test_use_type_in_kernel(self):
        self.skip_unsupported()

        @cuda.jit
        def kernel():
            bfloat16(3.14)

        kernel[1, 1]()

    def test_math_bindings(self):
        self.skip_unsupported()

        exp_functions = [math.exp]
        if sys.version_info >= (3, 11):
            exp_functions += [math.exp2]

        functions = [
            math.trunc,
            math.ceil,
            math.floor,
            math.sqrt,
            math.log,
            math.log10,
            math.cos,
            math.sin,
            math.tanh,
        ] + exp_functions

        for f in functions:
            with self.subTest(func=f):

                @cuda.jit
                def kernel(arr):
                    x = bfloat16(3.14)
                    y = f(x)
                    arr[0] = float32(y)

                arr = cuda.device_array((1,), dtype="float32")
                kernel[1, 1](arr)

                if f in exp_functions:
                    self.assertAlmostEqual(arr[0], f(3.14), delta=1e-1)
                else:
                    self.assertAlmostEqual(arr[0], f(3.14), delta=1e-2)

    def test_arithmetic_intrinsics_basic(self):
        self.skip_unsupported()

        @cuda.jit
        def kernel(out):
            a = bfloat16(1.25)
            b = bfloat16(-2.5)

            out[0] = float32(habs(b))
            out[1] = float32(hadd(a, b))
            out[2] = float32(hsub(a, b))
            out[3] = float32(hmul(a, b))
            out[4] = float32(hdiv(b, a))
            out[5] = float32(hneg(a))
            out[6] = float32(hfma(a, b, b))

            out[7] = float32(hadd_rn(a, b))
            out[8] = float32(hsub_rn(a, b))
            out[9] = float32(hmul_rn(a, b))

        out = cuda.device_array((10,), dtype="float32")
        kernel[1, 1](out)

        a = 1.25
        b = -2.5
        expected = [
            abs(b),
            a + b,
            a - b,
            a * b,
            b / a,
            -a,
            a * b + b,
            a + b,
            a - b,
            a * b,
        ]
        for i, exp in enumerate(expected):
            self.assertAlmostEqual(out[i], exp, delta=1e-2)

    def test_arithmetic_intrinsics_saturating(self):
        self.skip_unsupported()

        @cuda.jit
        def kernel(out):
            a = bfloat16(1.5)
            b = bfloat16(0.75)

            out[0] = float32(hadd_sat(a, b))  # 2.25 -> 1.0
            out[1] = float32(hsub_sat(b, a))  # -0.75 -> 0.0
            out[2] = float32(hmul_sat(a, b))  # 1.125 -> 1.0
            out[3] = float32(hfma_sat(a, b, a))  # 1.125 + 1.5 -> 1.0

        out = cuda.device_array((4,), dtype="float32")
        kernel[1, 1](out)

        self.assertAlmostEqual(out[0], 1.0, delta=1e-3)
        self.assertAlmostEqual(out[1], 0.0, delta=1e-3)
        self.assertAlmostEqual(out[2], 1.0, delta=1e-3)
        self.assertAlmostEqual(out[3], 1.0, delta=1e-3)

        # Also check they are clamped within [0, 1]
        for i in range(4):
            self.assertGreaterEqual(out[i], 0.0)
            self.assertLessEqual(out[i], 1.0)

    def test_fma_relu_intrinsic(self):
        self.skip_unsupported()

        @cuda.jit
        def kernel(out):
            a = bfloat16(-1.5)
            b = bfloat16(2.0)
            c = bfloat16(0.0)

            out[0] = float32(hfma_relu(a, b, c))  # -3.0 -> relu -> 0.0

        out = cuda.device_array((1,), dtype="float32")
        kernel[1, 1](out)

        self.assertAlmostEqual(out[0], 0.0, delta=1e-3)

    def test_comparison_intrinsics(self):
        self.skip_unsupported()

        def make_kernel(cmpfn):
            @cuda.jit
            def kernel(out, a, b):
                a_bf16 = bfloat16(a)
                b_bf16 = bfloat16(b)
                out[0] = cmpfn(a_bf16, b_bf16)

            return kernel

        comparisons = [heq, hne, hge, hgt, hle, hlt]
        ops = [
            lambda x, y: x == y,
            lambda x, y: x != y,
            lambda x, y: x >= y,
            lambda x, y: x > y,
            lambda x, y: x <= y,
            lambda x, y: x < y,
        ]

        for cmpfn, op in zip(comparisons, ops):
            with self.subTest(cmpfn=cmpfn):
                kernel = make_kernel(cmpfn)
                out = cuda.device_array((1,), dtype="bool")

                a = 3.0
                b = 3.0
                kernel[1, 1](out, a, b)
                self.assertEqual(bool(out[0]), op(3.0, 3.0))

                a = 3.0
                b = 4.0
                kernel[1, 1](out, a, b)
                self.assertEqual(bool(out[0]), op(3.0, 4.0))

                a = 4.0
                b = 3.0
                kernel[1, 1](out, a, b)
                self.assertEqual(bool(out[0]), op(4.0, 3.0))

    def test_hmax_hmin_intrinsics(self):
        self.skip_unsupported()

        @cuda.jit
        def kernel(out):
            a = bfloat16(3.0)
            b = bfloat16(4.0)
            out[0] = float32(hmax(a, b))
            out[1] = float32(hmin(a, b))

        out = cuda.device_array((2,), dtype="float32")
        kernel[1, 1](out)
        self.assertAlmostEqual(out[0], 4.0, delta=1e-3)
        self.assertAlmostEqual(out[1], 3.0, delta=1e-3)

    def test_nan_and_inf_intrinsics(self):
        self.skip_unsupported()

        @cuda.jit
        def kernel(out_bool, out_int):
            nanv = bfloat16(float("nan"))
            infv = bfloat16(float("inf"))
            out_bool[0] = hisnan(nanv)
            out_int[0] = hisinf(infv)

        out_bool = cuda.device_array((1,), dtype="bool")
        out_int = cuda.device_array((1,), dtype="int32")
        kernel[1, 1](out_bool, out_int)
        self.assertTrue(bool(out_bool[0]))
        self.assertNotEqual(int(out_int[0]), 0)

    def test_hmax_nan_hmin_nan_intrinsics(self):
        self.skip_unsupported()

        @cuda.jit
        def kernel(out):
            a = bfloat16(float("nan"))
            b = bfloat16(2.0)
            out[0] = float32(hmax_nan(a, b))
            out[1] = float32(hmin_nan(a, b))
            out[2] = float32(hmax(a, b))
            out[3] = float32(hmin(a, b))

        out = cuda.device_array((4,), dtype="float32")
        kernel[1, 1](out)
        # NaN-propagating variants should produce NaN
        self.assertTrue(math.isnan(out[0]))
        self.assertTrue(math.isnan(out[1]))
        # Non-NaN variants should return the non-NaN operand
        self.assertAlmostEqual(out[2], 2.0, delta=1e-3)
        self.assertAlmostEqual(out[3], 2.0, delta=1e-3)

    def test_bfloat16_as_bitcast(self):
        self.skip_unsupported()

        @cuda.jit
        def roundtrip_kernel(test_val, i2, u2):
            i2[0] = int16_as_bfloat16(bfloat16_as_int16(test_val))
            u2[0] = uint16_as_bfloat16(bfloat16_as_uint16(test_val))

        test_val = np.int16(0x3FC0)  # 1.5 in bfloat16
        i2 = cuda.device_array((1,), dtype="int16")
        u2 = cuda.device_array((1,), dtype="uint16")
        roundtrip_kernel[1, 1](test_val, i2, u2)

        self.assertEqual(i2[0], test_val)
        self.assertEqual(u2[0], test_val)

    def test_to_integer_conversions(self):
        self.skip_unsupported()

        @cuda.jit
        def kernel(test_val, i1, i2, i3, i4, u1, u2, u3, u4):
            a = int16_as_bfloat16(test_val)

            i1[0] = bfloat16_to_int8_rz(a)
            u1[0] = bfloat16_to_uint8_rz(a)
            i2[0] = bfloat16_to_int16_rn(a)
            i2[1] = bfloat16_to_int16_rz(a)
            i2[2] = bfloat16_to_int16_rd(a)
            i2[3] = bfloat16_to_int16_ru(a)
            u2[0] = bfloat16_to_uint16_rn(a)
            u2[1] = bfloat16_to_uint16_rz(a)
            u2[2] = bfloat16_to_uint16_rd(a)
            u2[3] = bfloat16_to_uint16_ru(a)
            i3[0] = bfloat16_to_int32_rn(a)
            i3[1] = bfloat16_to_int32_rz(a)
            i3[2] = bfloat16_to_int32_rd(a)
            i3[3] = bfloat16_to_int32_ru(a)
            u3[0] = bfloat16_to_uint32_rn(a)
            u3[1] = bfloat16_to_uint32_rz(a)
            u3[2] = bfloat16_to_uint32_rd(a)
            u3[3] = bfloat16_to_uint32_ru(a)
            i4[0] = bfloat16_to_int64_rn(a)
            i4[1] = bfloat16_to_int64_rz(a)
            i4[2] = bfloat16_to_int64_rd(a)
            i4[3] = bfloat16_to_int64_ru(a)
            u4[0] = bfloat16_to_uint64_rn(a)
            u4[1] = bfloat16_to_uint64_rz(a)
            u4[2] = bfloat16_to_uint64_rd(a)
            u4[3] = bfloat16_to_uint64_ru(a)

        # rz
        i1 = cuda.device_array((1,), dtype="int8")
        # rn, rz, rd, ru
        i2 = cuda.device_array((4,), dtype="int16")
        i3 = cuda.device_array((4,), dtype="int32")
        i4 = cuda.device_array((4,), dtype="int64")
        # rz
        u1 = cuda.device_array((1,), dtype="uint8")
        # rn, rz, rd, ru
        u2 = cuda.device_array((4,), dtype="uint16")
        u3 = cuda.device_array((4,), dtype="uint32")
        u4 = cuda.device_array((4,), dtype="uint64")

        test_val = np.int16(0x3FC0)  # 1.5 in bfloat16

        kernel[1, 1](test_val, i1, i2, i3, i4, u1, u2, u3, u4)

        self.assertEqual(i1[0], 1)
        self.assertEqual(u1[0], 1)

        np.testing.assert_equal(i2, np.array([2, 1, 1, 2], "int16"))
        np.testing.assert_equal(i3, np.array([2, 1, 1, 2], "int32"))
        np.testing.assert_equal(i4, np.array([2, 1, 1, 2], "int64"))
        np.testing.assert_equal(u2, np.array([2, 1, 1, 2], "uint16"))
        np.testing.assert_equal(u3, np.array([2, 1, 1, 2], "uint32"))
        np.testing.assert_equal(u4, np.array([2, 1, 1, 2], "uint64"))

    def test_from_integer_conversions(self):
        self.skip_unsupported()

        test_val = 789

        @cuda.jit
        def kernel(out):
            i2 = int16(test_val)
            i3 = int32(test_val)
            i4 = int64(test_val)
            u2 = uint16(test_val)
            u3 = uint32(test_val)
            u4 = uint64(test_val)

            i2rn = int16_to_bfloat16_rn(i2)
            i2rz = int16_to_bfloat16_rz(i2)
            i2rd = int16_to_bfloat16_rd(i2)
            i2ru = int16_to_bfloat16_ru(i2)

            u2rn = uint16_to_bfloat16_rn(u2)
            u2rz = uint16_to_bfloat16_rz(u2)
            u2rd = uint16_to_bfloat16_rd(u2)
            u2ru = uint16_to_bfloat16_ru(u2)

            i3rn = int32_to_bfloat16_rn(i3)
            i3rz = int32_to_bfloat16_rz(i3)
            i3rd = int32_to_bfloat16_rd(i3)
            i3ru = int32_to_bfloat16_ru(i3)

            u3rn = uint32_to_bfloat16_rn(u3)
            u3rz = uint32_to_bfloat16_rz(u3)
            u3rd = uint32_to_bfloat16_rd(u3)
            u3ru = uint32_to_bfloat16_ru(u3)

            i4rn = int64_to_bfloat16_rn(i4)
            i4rz = int64_to_bfloat16_rz(i4)
            i4rd = int64_to_bfloat16_rd(i4)
            i4ru = int64_to_bfloat16_ru(i4)

            u4rn = uint64_to_bfloat16_rn(u4)
            u4rz = uint64_to_bfloat16_rz(u4)
            u4rd = uint64_to_bfloat16_rd(u4)
            u4ru = uint64_to_bfloat16_ru(u4)

            out[0] = bfloat16_as_int16(i2rn)
            out[1] = bfloat16_as_int16(i2rz)
            out[2] = bfloat16_as_int16(i2rd)
            out[3] = bfloat16_as_int16(i2ru)
            out[4] = bfloat16_as_int16(u2rn)
            out[5] = bfloat16_as_int16(u2rz)
            out[6] = bfloat16_as_int16(u2rd)
            out[7] = bfloat16_as_int16(u2ru)
            out[8] = bfloat16_as_int16(i3rn)
            out[9] = bfloat16_as_int16(i3rz)
            out[10] = bfloat16_as_int16(i3rd)
            out[11] = bfloat16_as_int16(i3ru)
            out[12] = bfloat16_as_int16(u3rn)
            out[13] = bfloat16_as_int16(u3rz)
            out[14] = bfloat16_as_int16(u3rd)
            out[15] = bfloat16_as_int16(u3ru)
            out[16] = bfloat16_as_int16(i4rn)
            out[17] = bfloat16_as_int16(i4rz)
            out[18] = bfloat16_as_int16(i4rd)
            out[19] = bfloat16_as_int16(i4ru)
            out[20] = bfloat16_as_int16(u4rn)
            out[21] = bfloat16_as_int16(u4rz)
            out[22] = bfloat16_as_int16(u4rd)
            out[23] = bfloat16_as_int16(u4ru)

        out = cuda.device_array((24,), dtype="int16")
        kernel[1, 1](out)
        res = out.copy_to_host()

        i2 = np.int16(789).astype(mldtypes_bf16).view("int16")
        i3 = np.int32(789).astype(mldtypes_bf16).view("int16")
        i4 = np.int64(789).astype(mldtypes_bf16).view("int16")
        u2 = np.uint16(789).astype(mldtypes_bf16).view("int16")
        u3 = np.uint32(789).astype(mldtypes_bf16).view("int16")
        u4 = np.uint64(789).astype(mldtypes_bf16).view("int16")

        i2arr = np.array([i2] * 4)
        i3arr = np.array([i3] * 4)
        i4arr = np.array([i4] * 4)
        u2arr = np.array([u2] * 4)
        u3arr = np.array([u3] * 4)
        u4arr = np.array([u4] * 4)

        two = np.ones_like(res[0:4]) * 2
        np.testing.assert_array_less(_bf16_ulp_distance(res[0:4], i2arr), two)
        np.testing.assert_array_less(_bf16_ulp_distance(res[4:8], i3arr), two)
        np.testing.assert_array_less(_bf16_ulp_distance(res[8:12], i4arr), two)
        np.testing.assert_array_less(_bf16_ulp_distance(res[12:16], u2arr), two)
        np.testing.assert_array_less(_bf16_ulp_distance(res[16:20], u3arr), two)
        np.testing.assert_array_less(_bf16_ulp_distance(res[20:24], u4arr), two)

    def test_to_float_conversions(self):
        self.skip_unsupported()

        @cuda.jit
        def kernel(out):
            a = bfloat16(1.5)
            out[0] = bfloat16_to_float32(a)

        out = cuda.device_array((1,), dtype="float32")
        kernel[1, 1](out)

        self.assertAlmostEqual(out[0], 1.5, delta=1e-7)  # conversion is exact

    def test_from_float_conversions(self):
        self.skip_unsupported()

        test_val = 1.5

        @cuda.jit
        def kernel(out):
            f4 = float32(test_val)
            f8 = float64(test_val)

            f4rn = float32_to_bfloat16_rn(f4)
            f4rz = float32_to_bfloat16_rz(f4)
            f4rd = float32_to_bfloat16_rd(f4)
            f4ru = float32_to_bfloat16_ru(f4)

            f4_default = float32_to_bfloat16(f4)
            f8_default = float64_to_bfloat16(f8)

            out[0] = bfloat16_as_int16(f4rn)
            out[1] = bfloat16_as_int16(f4rz)
            out[2] = bfloat16_as_int16(f4rd)
            out[3] = bfloat16_as_int16(f4ru)
            out[4] = bfloat16_as_int16(f4_default)
            out[5] = bfloat16_as_int16(f8_default)

        out = cuda.device_array((6,), dtype="int16")
        kernel[1, 1](out)
        raw = out.copy_to_host()

        f4_expected = (
            np.array([test_val] * 4, "float32")
            .astype(mldtypes_bf16)
            .view("int16")
        )
        f8_expected = (
            np.array([test_val] * 1, "float64")
            .astype(mldtypes_bf16)
            .view("int16")
        )

        np.testing.assert_array_less(
            _bf16_ulp_distance(raw[0:4], f4_expected), 2
        )
        np.testing.assert_array_less(
            _bf16_ulp_distance(raw[4:], f8_expected), 2
        )

    def test_bfloat16_type_import(self):
        self.skip_unsupported()


def _bf16_ulp_rank(bits_int16: np.ndarray) -> np.ndarray:
    """
    Compute the ULP rank of a bfloat16 value. Input is the bits of the bfloat16 value as an int16.
    The ULP rank is the number of ULPs between the value and 0.
    Negative values are performed the inverse of 2's complement before computing the rank.
    """
    u = bits_int16.view(np.uint16)
    sign = u >> 15
    return np.where(sign == 0, u + 0x8000, 0x8000 - u).astype(np.int32)


def _bf16_ulp_distance(
    a_bits_int16: np.ndarray, b_bits_int16: np.ndarray
) -> np.ndarray:
    """
    Compute the difference between two bfloat16 values in ULPs.
    """
    return np.abs(_bf16_ulp_rank(a_bits_int16) - _bf16_ulp_rank(b_bits_int16))
